from typing import Tuple

import mlx.core as mx
import mlx.nn as nn

from .attention import MultiHeadedAttention, RelPositionMultiHeadedAttention
from .convolution import ConvolutionModule
from .embedding import EspnetRelPositionalEncoding, RelPositionalEncoding
from .encoder_layer import ConformerEncoderLayer
from .positionwise_feed_forward import PositionwiseFeedForward
from .subsampling import LinearNoSubsampling


class Upsample1D(nn.Module):
    """A 1D upsampling layer with an optional convolution.

    Parameters:
        channels (int): number of channels in the inputs and outputs.
        out_channels (int): number of output channels.
        stride (int): upsampling factor. Defaults to 2.
    """

    def __init__(self, channels: int, out_channels: int, stride: int = 2):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels
        self.stride = stride
        # In this mode, first repeat interpolate, then conv with stride=1
        self.conv = nn.Conv1d(
            self.channels, self.out_channels, stride * 2 + 1, stride=1, padding=0
        )

    def __call__(
        self, inputs: mx.array, input_lengths: mx.array
    ) -> Tuple[mx.array, mx.array]:
        """
        Args:
            inputs: Input tensor (B, C, T) - PyTorch format
            input_lengths: Lengths tensor (B,)

        Returns:
            outputs: Upsampled tensor (B, C, T*stride) - PyTorch format
            output_lengths: Updated lengths (B,)
        """
        # Upsample using nearest neighbor interpolation
        B, C, T = inputs.shape
        new_T = T * self.stride

        # Repeat each timestep stride times
        outputs = mx.repeat(inputs, self.stride, axis=2)

        # Pad on the left (axis 2 in PyTorch format B,C,T)
        outputs = mx.pad(outputs, [(0, 0), (0, 0), (self.stride * 2, 0)])

        # Transpose to MLX format (B, T, C) for conv
        outputs = mx.transpose(outputs, [0, 2, 1])

        # Apply convolution (MLX Conv1d expects B, T, C)
        outputs = self.conv(outputs)

        # Transpose back to PyTorch format (B, C, T)
        outputs = mx.transpose(outputs, [0, 2, 1])

        return outputs, input_lengths * self.stride


class PreLookaheadLayer(nn.Module):
    """Pre-lookahead layer for causal processing."""

    def __init__(self, channels: int, pre_lookahead_len: int = 1):
        super().__init__()
        self.channels = channels
        self.pre_lookahead_len = pre_lookahead_len
        self.conv1 = nn.Conv1d(
            channels,
            channels,
            kernel_size=pre_lookahead_len + 1,
            stride=1,
            padding=0,
        )
        self.conv2 = nn.Conv1d(
            channels,
            channels,
            kernel_size=3,
            stride=1,
            padding=0,
        )

    def __call__(self, inputs: mx.array, context: mx.array = None) -> mx.array:
        """
        Args:
            inputs: Input tensor (B, T, C)
            context: Optional context tensor for lookahead (B, pre_lookahead_len, C)
                     Used during streaming inference to provide future context.

        Returns:
            outputs: Output tensor (B, T, C)
        """
        # Input is (B, T, C) - MLX Conv1d expects (B, T, C)
        outputs = inputs

        # Look ahead padding on time dimension (axis 1)
        if context is None or context.shape[1] == 0:
            # No context - pad with zeros
            outputs = mx.pad(outputs, [(0, 0), (0, self.pre_lookahead_len), (0, 0)])
        else:
            # Use context for lookahead
            assert (
                context.shape[1] == self.pre_lookahead_len
            ), f"Context length {context.shape[1]} != pre_lookahead_len {self.pre_lookahead_len}"
            outputs = mx.concatenate([outputs, context], axis=1)
            # Pad remaining if needed
            remaining = self.pre_lookahead_len - context.shape[1]
            if remaining > 0:
                outputs = mx.pad(outputs, [(0, 0), (0, remaining), (0, 0)])

        outputs = nn.leaky_relu(self.conv1(outputs))

        # Output padding on time dimension (axis 1)
        outputs = mx.pad(outputs, [(0, 0), (2, 0), (0, 0)])
        outputs = self.conv2(outputs)

        # Residual connection
        outputs = outputs + inputs
        return outputs


def make_pad_mask(lengths: mx.array, max_len: int = 0) -> mx.array:
    """Make mask tensor containing indices of padded part.

    Args:
        lengths: Batch of lengths (B,).
        max_len: Maximum length. If 0, uses max of lengths.

    Returns:
        Mask tensor containing indices of padded part (B, T).

    Examples:
        >>> lengths = [5, 3, 2]
        >>> make_pad_mask(lengths)
        masks = [[0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 1],
                 [0, 0, 1, 1, 1]]
    """
    batch_size = lengths.shape[0]
    max_len = max_len if max_len > 0 else int(mx.max(lengths).item())

    seq_range = mx.arange(max_len)
    seq_range_expand = mx.expand_dims(seq_range, 0)  # (1, T)
    seq_range_expand = mx.broadcast_to(
        seq_range_expand, (batch_size, max_len)
    )  # (B, T)

    seq_length_expand = mx.expand_dims(lengths, -1)  # (B, 1)
    mask = seq_range_expand >= seq_length_expand
    return mask


def subsequent_chunk_mask(
    size: int,
    chunk_size: int,
    num_left_chunks: int = -1,
) -> mx.array:
    """Create mask for subsequent steps (size, size) with chunk size.

    This is for streaming encoder.

    Args:
        size: size of mask
        chunk_size: size of chunk
        num_left_chunks: number of left chunks
            <0: use full chunk
            >=0: use num_left_chunks

    Returns:
        Mask tensor (size, size)

    Examples:
        >>> subsequent_chunk_mask(4, 2)
        [[1, 1, 0, 0],
         [1, 1, 0, 0],
         [1, 1, 1, 1],
         [1, 1, 1, 1]]
    """
    pos_idx = mx.arange(size)
    block_value = ((pos_idx // chunk_size) + 1) * chunk_size
    ret = mx.expand_dims(pos_idx, 0) < mx.expand_dims(block_value, 1)
    return ret


def add_optional_chunk_mask(
    xs: mx.array,
    masks: mx.array,
    use_dynamic_chunk: bool,
    use_dynamic_left_chunk: bool,
    decoding_chunk_size: int,
    static_chunk_size: int,
    num_decoding_left_chunks: int,
    enable_full_context: bool = True,
) -> mx.array:
    """Apply optional mask for encoder.

    Args:
        xs: padded input, (B, L, D), L for max length
        masks: mask for xs, (B, 1, L)
        use_dynamic_chunk: whether to use dynamic chunk or not
        use_dynamic_left_chunk: whether to use dynamic left chunk for training.
        decoding_chunk_size: decoding chunk size for dynamic chunk, it's
            0: default for training, use random dynamic chunk.
            <0: for decoding, use full chunk.
            >0: for decoding, use fixed chunk size as set.
        static_chunk_size: chunk size for static chunk training/decoding
        num_decoding_left_chunks: number of left chunks
            >=0: use num_decoding_left_chunks
            <0: use all left chunks
        enable_full_context:
            True: chunk size is either [1, 25] or full context(max_len)
            False: chunk size ~ U[1, 25]

    Returns:
        chunk mask of the input xs.
    """
    # Whether to use chunk mask or not
    if use_dynamic_chunk:
        max_len = xs.shape[1]
        if decoding_chunk_size < 0:
            chunk_size = max_len
            num_left_chunks = -1
        elif decoding_chunk_size > 0:
            chunk_size = decoding_chunk_size
            num_left_chunks = num_decoding_left_chunks
        else:
            # For training, use random chunk size
            import random

            chunk_size = random.randint(1, max_len - 1)
            num_left_chunks = -1
            if chunk_size > max_len // 2 and enable_full_context:
                chunk_size = max_len
            else:
                chunk_size = chunk_size % 25 + 1
                if use_dynamic_left_chunk:
                    max_left_chunks = (max_len - 1) // chunk_size
                    num_left_chunks = random.randint(0, max_left_chunks)

        chunk_masks = subsequent_chunk_mask(xs.shape[1], chunk_size, num_left_chunks)
        chunk_masks = mx.expand_dims(chunk_masks, 0)  # (1, L, L)
        chunk_masks = masks & chunk_masks  # (B, L, L)
    elif static_chunk_size > 0:
        num_left_chunks = num_decoding_left_chunks
        chunk_masks = subsequent_chunk_mask(
            xs.shape[1], static_chunk_size, num_left_chunks
        )
        chunk_masks = mx.expand_dims(chunk_masks, 0)  # (1, L, L)
        chunk_masks = masks & chunk_masks  # (B, L, L)
    else:
        chunk_masks = masks

    # Check for all-false masks and fix them
    mask_sums = mx.sum(chunk_masks, axis=-1)
    if mx.any(mask_sums == 0):
        # Force set to true where all false
        chunk_masks = mx.where(
            mx.expand_dims(mask_sums == 0, -1), mx.ones_like(chunk_masks), chunk_masks
        )

    return chunk_masks


class UpsampleConformerEncoder(nn.Module):
    """Conformer encoder with upsampling for speech synthesis."""

    def __init__(
        self,
        input_size: int = 512,
        output_size: int = 512,
        attention_heads: int = 8,
        linear_units: int = 2048,
        num_blocks: int = 6,
        num_up_blocks: int = 4,
        dropout_rate: float = 0.1,
        positional_dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        input_layer: str = "linear",
        pos_enc_layer_type: str = "rel_pos_espnet",
        normalize_before: bool = True,
        static_chunk_size: int = 0,
        use_dynamic_chunk: bool = False,
        use_dynamic_left_chunk: bool = False,
        positionwise_conv_kernel_size: int = 1,
        macaron_style: bool = False,
        selfattention_layer_type: str = "rel_selfattn",
        activation_type: str = "swish",
        use_cnn_module: bool = False,
        cnn_module_kernel: int = 15,
        causal: bool = False,
        cnn_module_norm: str = "batch_norm",
        key_bias: bool = True,
        pre_lookahead_len: int = 3,
        upsample_stride: int = 2,
    ):
        """
        Args:
            input_size: input dimension
            output_size: dimension of attention
            attention_heads: the number of heads of multi head attention
            linear_units: the hidden units number of position-wise feed forward
            num_blocks: the number of encoder blocks before upsampling
            num_up_blocks: the number of encoder blocks after upsampling
            dropout_rate: dropout rate
            attention_dropout_rate: dropout rate in attention
            positional_dropout_rate: dropout rate after adding positional encoding
            input_layer: input layer type (currently only "linear" supported)
            pos_enc_layer_type: positional encoding layer type
            normalize_before: use layer_norm before each sub-block
            static_chunk_size: chunk size for static chunk training/decoding
            use_dynamic_chunk: whether use dynamic chunk size for training
            use_dynamic_left_chunk: whether use dynamic left chunk
            macaron_style: whether to use macaron style FFN
            selfattention_layer_type: self-attention layer type
            activation_type: activation function type
            use_cnn_module: whether to use convolution module
            cnn_module_kernel: kernel size of convolution module
            causal: whether to use causal convolution
            cnn_module_norm: normalization type for convolution module
            key_bias: whether use bias in attention.linear_k
            pre_lookahead_len: length of pre-lookahead window for streaming
            upsample_stride: stride for temporal upsampling (2 = 2x upsample)
        """
        super().__init__()
        self._output_size = output_size

        # Select positional encoding class based on pos_enc_layer_type
        if pos_enc_layer_type == "rel_pos_espnet":
            pos_enc_class = EspnetRelPositionalEncoding(
                output_size, positional_dropout_rate
            )
        elif pos_enc_layer_type == "rel_pos":
            pos_enc_class = RelPositionalEncoding(output_size, positional_dropout_rate)
        else:
            raise ValueError(f"Unsupported pos_enc_layer_type: {pos_enc_layer_type}")

        # Input embedding layer
        if input_layer == "linear":
            self.embed = LinearNoSubsampling(
                input_size,
                output_size,
                dropout_rate,
                pos_enc_class,
            )
        else:
            raise ValueError(f"Unsupported input_layer: {input_layer}")

        self.normalize_before = normalize_before
        self.after_norm = nn.LayerNorm(output_size, eps=1e-5)
        self.static_chunk_size = static_chunk_size
        self.use_dynamic_chunk = use_dynamic_chunk
        self.use_dynamic_left_chunk = use_dynamic_left_chunk

        # Activation function
        if activation_type == "swish":
            activation = nn.SiLU()
        else:
            raise ValueError(f"Unsupported activation: {activation_type}")

        # Self-attention layer
        if selfattention_layer_type == "rel_selfattn":
            self_attn_class = RelPositionMultiHeadedAttention
        elif selfattention_layer_type == "selfattn":
            self_attn_class = MultiHeadedAttention
        else:
            raise ValueError(
                f"Unsupported selfattention_layer_type: {selfattention_layer_type}"
            )

        # Pre-lookahead layer
        self.pre_lookahead_layer = PreLookaheadLayer(
            channels=output_size, pre_lookahead_len=pre_lookahead_len
        )

        # Main encoder layers - use indexed attributes for weight loading
        self._num_encoders = num_blocks
        for i in range(num_blocks):
            layer = ConformerEncoderLayer(
                output_size,
                self_attn_class(
                    attention_heads,
                    output_size,
                    attention_dropout_rate,
                    key_bias,
                ),
                PositionwiseFeedForward(
                    output_size, linear_units, dropout_rate, activation
                ),
                (
                    PositionwiseFeedForward(
                        output_size, linear_units, dropout_rate, activation
                    )
                    if macaron_style
                    else None
                ),
                (
                    ConvolutionModule(
                        output_size,
                        cnn_module_kernel,
                        activation,
                        cnn_module_norm,
                        causal,
                    )
                    if use_cnn_module
                    else None
                ),
                dropout_rate,
                normalize_before,
            )
            setattr(self, f"encoders_{i}", layer)

        # Upsampling layer
        self.upsample_stride = upsample_stride
        self.up_layer = Upsample1D(
            channels=output_size, out_channels=output_size, stride=upsample_stride
        )

        # Upsampling embedding layer (uses same pos_enc_layer_type)
        if pos_enc_layer_type == "rel_pos_espnet":
            up_pos_enc_class = EspnetRelPositionalEncoding(
                output_size, positional_dropout_rate
            )
        elif pos_enc_layer_type == "rel_pos":
            up_pos_enc_class = RelPositionalEncoding(
                output_size, positional_dropout_rate
            )
        else:
            raise ValueError(f"Unsupported pos_enc_layer_type: {pos_enc_layer_type}")

        if input_layer == "linear":
            self.up_embed = LinearNoSubsampling(
                input_size,
                output_size,
                dropout_rate,
                up_pos_enc_class,
            )
        else:
            raise ValueError(f"Unsupported input_layer: {input_layer}")

        # Upsampling encoder layers - use indexed attributes for weight loading
        self._num_up_encoders = num_up_blocks
        for i in range(num_up_blocks):
            layer = ConformerEncoderLayer(
                output_size,
                self_attn_class(
                    attention_heads,
                    output_size,
                    attention_dropout_rate,
                    key_bias,
                ),
                PositionwiseFeedForward(
                    output_size, linear_units, dropout_rate, activation
                ),
                (
                    PositionwiseFeedForward(
                        output_size, linear_units, dropout_rate, activation
                    )
                    if macaron_style
                    else None
                ),
                (
                    ConvolutionModule(
                        output_size,
                        cnn_module_kernel,
                        activation,
                        cnn_module_norm,
                        causal,
                    )
                    if use_cnn_module
                    else None
                ),
                dropout_rate,
                normalize_before,
            )
            setattr(self, f"up_encoders_{i}", layer)

    def output_size(self) -> int:
        return self._output_size

    def __call__(
        self,
        xs: mx.array,
        xs_lens: mx.array,
        context: mx.array = None,
        decoding_chunk_size: int = 0,
        num_decoding_left_chunks: int = -1,
        streaming: bool = False,
    ) -> Tuple[mx.array, mx.array]:
        """Embed positions in tensor.

        Args:
            xs: padded input tensor (B, T, D)
            xs_lens: input length (B)
            context: optional context tensor for lookahead (B, pre_lookahead_len, D)
                     Used during streaming inference to provide future context.
            decoding_chunk_size: decoding chunk size for dynamic chunk
                0: default for training, use random dynamic chunk.
                <0: for decoding, use full chunk.
                >0: for decoding, use fixed chunk size as set.
            num_decoding_left_chunks: number of left chunks
                >=0: use num_decoding_left_chunks
                <0: use all left chunks
            streaming: whether to use streaming (chunk-based) attention
                When True, uses static_chunk_size for causal masking
                When False, uses full context attention (chunk_size=0)

        Returns:
            encoder output tensor xs, and subsampled masks
            xs: padded output tensor (B, T' ~= T*2, D)
            masks: batch padding mask (B, 1, T' ~= T*2)
        """
        T = xs.shape[1]
        masks = mx.logical_not(make_pad_mask(xs_lens, T))
        masks = mx.expand_dims(masks, 1)  # (B, 1, T)

        xs, pos_emb, masks = self.embed(xs, masks)

        # Embed context if provided
        embedded_context = None
        if context is not None and context.shape[1] > 0:
            context_masks = mx.ones((1, 1, context.shape[1]), dtype=mx.bool_)
            embedded_context, _, _ = self.embed(
                context, context_masks, offset=xs.shape[1]
            )

        mask_pad = masks  # (B, 1, T)

        # Use static_chunk_size when streaming, otherwise full context (0)
        effective_chunk_size = self.static_chunk_size if streaming else 0

        chunk_masks = add_optional_chunk_mask(
            xs,
            masks,
            self.use_dynamic_chunk,
            self.use_dynamic_left_chunk,
            decoding_chunk_size,
            effective_chunk_size,
            num_decoding_left_chunks,
        )

        # Lookahead + conformer encoder
        xs = self.pre_lookahead_layer(xs, context=embedded_context)
        xs = self.forward_layers(xs, chunk_masks, pos_emb, mask_pad)

        # Upsample + conformer encoder
        xs = mx.transpose(xs, [0, 2, 1])  # (B, D, T)
        xs, xs_lens = self.up_layer(xs, xs_lens)
        xs = mx.transpose(xs, [0, 2, 1])  # (B, T', D)

        T = xs.shape[1]
        masks = mx.logical_not(make_pad_mask(xs_lens, T))
        masks = mx.expand_dims(masks, 1)  # (B, 1, T')

        xs, pos_emb, masks = self.up_embed(xs, masks)
        mask_pad = masks  # (B, 1, T')

        # Scale chunk size by upsample stride for upsampled encoder
        effective_up_chunk_size = effective_chunk_size * self.up_layer.stride

        chunk_masks = add_optional_chunk_mask(
            xs,
            masks,
            self.use_dynamic_chunk,
            self.use_dynamic_left_chunk,
            decoding_chunk_size,
            effective_up_chunk_size,
            num_decoding_left_chunks,
        )

        xs = self.forward_up_layers(xs, chunk_masks, pos_emb, mask_pad)

        if self.normalize_before:
            xs = self.after_norm(xs)

        return xs, masks

    @property
    def encoders(self):
        """Get encoder layers as a list."""
        return [getattr(self, f"encoders_{i}") for i in range(self._num_encoders)]

    @property
    def up_encoders(self):
        """Get up_encoder layers as a list."""
        return [getattr(self, f"up_encoders_{i}") for i in range(self._num_up_encoders)]

    def forward_layers(
        self, xs: mx.array, chunk_masks: mx.array, pos_emb: mx.array, mask_pad: mx.array
    ) -> mx.array:
        """Forward through main encoder layers."""
        for i in range(self._num_encoders):
            layer = getattr(self, f"encoders_{i}")
            xs, chunk_masks, _, _ = layer(xs, chunk_masks, pos_emb, mask_pad)
        return xs

    def forward_up_layers(
        self, xs: mx.array, chunk_masks: mx.array, pos_emb: mx.array, mask_pad: mx.array
    ) -> mx.array:
        """Forward through upsampling encoder layers."""
        for i in range(self._num_up_encoders):
            layer = getattr(self, f"up_encoders_{i}")
            xs, chunk_masks, _, _ = layer(xs, chunk_masks, pos_emb, mask_pad)
        return xs



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from .config import (
    AcousticTokenizerConfig,
    ModelConfig,
    Qwen2Config,
    SemanticTokenizerConfig,
)
from .vibevoice_asr import Model

__all__ = [
    "Model",
    "ModelConfig",
    "AcousticTokenizerConfig",
    "SemanticTokenizerConfig",
    "Qwen2Config",
]



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn


class EspnetRelPositionalEncoding(nn.Module):
    """Relative positional encoding module (ESPnet style).

    Creates positional encodings for both positive and negative relative positions.
    The output has length 2*T-1 for input of length T.

    See: https://arxiv.org/abs/1901.02860
    """

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.d_model = d_model
        self.xscale = float(d_model) ** 0.5
        self.max_len = max_len
        self.pe = None
        # Initialize with a reasonable size
        self._extend_pe(max_len)

    def _extend_pe(self, size: int):
        """Create or extend positional encoding matrix."""
        if self.pe is not None and self.pe.shape[1] >= size * 2 - 1:
            return

        # Create positive and negative position encodings
        position = mx.arange(0, size, dtype=mx.float32)[:, None]
        div_term = mx.exp(
            mx.arange(0, self.d_model, 2, dtype=mx.float32)
            * (-float(mx.log(mx.array(10000.0))) / self.d_model)
        )

        # Positive positions
        pe_positive = mx.zeros((size, self.d_model))
        pe_positive_sin = mx.sin(position * div_term)
        pe_positive_cos = mx.cos(position * div_term)
        # Interleave: pe[:, 0::2] = sin, pe[:, 1::2] = cos
        pe_positive = mx.concatenate(
            [pe_positive_sin[:, :, None], pe_positive_cos[:, :, None]], axis=-1
        ).reshape(size, self.d_model)

        # Negative positions
        pe_negative_sin = mx.sin(-position * div_term)
        pe_negative_cos = mx.cos(-position * div_term)
        pe_negative = mx.concatenate(
            [pe_negative_sin[:, :, None], pe_negative_cos[:, :, None]], axis=-1
        ).reshape(size, self.d_model)

        # Flip positive and concatenate: [pos_reversed, neg[1:]]
        pe_positive_flipped = pe_positive[::-1, :]  # Reverse order
        pe_negative_tail = pe_negative[1:, :]  # Skip position 0

        pe = mx.concatenate([pe_positive_flipped, pe_negative_tail], axis=0)
        self.pe = pe[None, :, :]  # (1, 2*size-1, d_model)

    def __call__(self, x: mx.array, offset: int = 0) -> Tuple[mx.array, mx.array]:
        """
        Args:
            x: Input tensor (B, T, D)
            offset: Position offset (unused, for compatibility)

        Returns:
            x: Scaled input (B, T, D)
            pos_emb: Positional encoding (1, 2*T-1, D)
        """
        T = x.shape[1]
        self._extend_pe(T)

        # Scale input by sqrt(d_model)
        x = x * self.xscale

        # Get positional embedding centered around current sequence
        # pe has shape (1, 2*max_len-1, d_model)
        # We want positions from (max_len - T) to (max_len + T - 1), which is 2*T-1 positions
        center = self.pe.shape[1] // 2
        pos_emb = self.pe[:, center - T + 1 : center + T, :]

        return x, pos_emb


class LinearInput(nn.Module):
    """Linear input projection with LayerNorm and positional encoding.

    Matches PyTorch LinearNoSubsampling with EspnetRelPositionalEncoding.
    """

    def __init__(self, input_size: int, output_size: int, dropout: float = 0.1):
        super().__init__()
        self.linear = nn.Linear(input_size, output_size)
        self.norm = nn.LayerNorm(output_size, eps=1e-5)
        self.pos_enc = EspnetRelPositionalEncoding(output_size, dropout=dropout)

    def __call__(
        self, x: mx.array, mask: mx.array
    ) -> Tuple[mx.array, mx.array, mx.array]:
        """
        Args:
            x: Input (B, T, D)
            mask: Mask (B, 1, T)

        Returns:
            x: Projected and scaled input (B, T, D)
            pos_emb: Positional encoding (1, 2*T-1, D)
            mask: Unchanged mask
        """
        x = self.linear(x)
        x = self.norm(x)
        x, pos_emb = self.pos_enc(x)  # pos_emb has shape (1, 2*T-1, D)
        return x, pos_emb, mask


class RelPositionMultiHeadedAttention(nn.Module):
    """Multi-head self-attention with relative positional encoding.

    Matches PyTorch RelPositionMultiHeadedAttention from ESPnet.
    See: https://arxiv.org/abs/1901.02860
    """

    def __init__(
        self,
        n_head: int,
        n_feat: int,
        dropout_rate: float = 0.0,
        key_bias: bool = True,
    ):
        super().__init__()
        self.n_head = n_head
        self.d_k = n_feat // n_head
        self.scale = self.d_k**-0.5

        self.linear_q = nn.Linear(n_feat, n_feat)
        self.linear_k = nn.Linear(n_feat, n_feat, bias=key_bias)
        self.linear_v = nn.Linear(n_feat, n_feat)
        self.linear_out = nn.Linear(n_feat, n_feat)

        # Relative positional encoding components
        self.linear_pos = nn.Linear(n_feat, n_feat, bias=False)
        # These are learnable parameters (initialized from weights)
        self.pos_bias_u = mx.zeros((n_head, self.d_k))
        self.pos_bias_v = mx.zeros((n_head, self.d_k))

    def _rel_shift(self, x: mx.array) -> mx.array:
        """Compute relative positional encoding shift.

        Args:
            x: Input tensor (batch, head, time1, 2*time1-1)

        Returns:
            Output tensor (batch, head, time1, time1)
        """
        B, n_head, T1, T2 = x.shape
        # Pad with zeros on the left
        zero_pad = mx.zeros((B, n_head, T1, 1))
        x_padded = mx.concatenate([zero_pad, x], axis=-1)

        # Reshape and extract the valid part
        x_padded = x_padded.reshape(B, n_head, T2 + 1, T1)
        x = x_padded[:, :, 1:, :]  # Remove first row
        x = x.reshape(B, n_head, T1, T2)

        # Keep only positions 0 to time1 (the valid relative positions)
        x = x[:, :, :, : T2 // 2 + 1]
        return x

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        pos_emb: Optional[mx.array] = None,
    ) -> mx.array:
        """
        Args:
            x: Input (B, T, D)
            mask: Attention mask (B, T) or (B, T, T)
            pos_emb: Positional embedding (1, 2*T-1, D)

        Returns:
            Output (B, T, D)
        """
        B, T, D = x.shape

        # Compute Q, K, V
        q = self.linear_q(x).reshape(B, T, self.n_head, self.d_k)
        k = self.linear_k(x).reshape(B, T, self.n_head, self.d_k).transpose(0, 2, 1, 3)
        v = self.linear_v(x).reshape(B, T, self.n_head, self.d_k).transpose(0, 2, 1, 3)

        # q stays as (B, T, n_head, d_k) for adding bias, then transpose
        # Content attention: (q + pos_bias_u) @ k^T
        q_with_bias_u = (q + self.pos_bias_u).transpose(
            0, 2, 1, 3
        )  # (B, n_head, T, d_k)
        matrix_ac = q_with_bias_u @ k.transpose(0, 1, 3, 2)  # (B, n_head, T, T)

        # Position attention: (q + pos_bias_v) @ p^T
        if pos_emb is not None:
            T_pos = pos_emb.shape[1]  # 2*T-1
            p = self.linear_pos(pos_emb)  # (1, 2*T-1, D)
            p = p.reshape(1, T_pos, self.n_head, self.d_k).transpose(
                0, 2, 1, 3
            )  # (1, n_head, 2*T-1, d_k)

            q_with_bias_v = (q + self.pos_bias_v).transpose(
                0, 2, 1, 3
            )  # (B, n_head, T, d_k)
            matrix_bd = q_with_bias_v @ p.transpose(0, 1, 3, 2)  # (B, n_head, T, 2*T-1)

            # Apply relative shift when shapes don't match
            if matrix_ac.shape != matrix_bd.shape:
                matrix_bd = self._rel_shift(matrix_bd)

            scores = (matrix_ac + matrix_bd) * self.scale
        else:
            scores = matrix_ac * self.scale

        # Apply mask if provided
        if mask is not None:
            if mask.ndim == 2:
                # (B, T) -> (B, 1, 1, T)
                mask_expanded = mask[:, None, None, :]
            else:
                # (B, T, T) -> (B, 1, T, T)
                mask_expanded = mask[:, None, :, :]
            # PyTorch uses mask.eq(0) and fills with -inf, we use inverted mask
            scores = mx.where(mask_expanded > 0, scores, mx.array(-float("inf")))

        attn = mx.softmax(scores, axis=-1)
        # Replace NaN from softmax(-inf) with 0
        attn = mx.where(mx.isnan(attn), mx.array(0.0), attn)

        out = attn @ v  # (B, n_head, T, d_k)

        # Reshape back
        out = out.transpose(0, 2, 1, 3).reshape(B, T, D)
        return self.linear_out(out)


class PositionwiseFeedForward(nn.Module):
    """Position-wise feed-forward network.

    Matches PyTorch PositionwiseFeedForward from ESPnet/CosyVoice.
    Uses Swish (SiLU) activation as configured in the original model.
    """

    def __init__(self, d_model: int, d_inner: int, dropout: float = 0.1):
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_inner)
        self.w_2 = nn.Linear(d_inner, d_model)

    def __call__(self, x: mx.array) -> mx.array:
        # Use SiLU (Swish) activation to match PyTorch
        return self.w_2(nn.silu(self.w_1(x)))


class ConformerEncoderLayer(nn.Module):
    """Single Conformer encoder layer.

    Matches PyTorch ConformerEncoderLayer from ESPnet/CosyVoice.
    Uses pre-norm (normalize_before=True) style.
    """

    def __init__(
        self,
        size: int,
        n_head: int,
        d_inner: int,
        dropout_rate: float = 0.1,
        key_bias: bool = True,
    ):
        super().__init__()
        # Use eps=1e-12 to match PyTorch
        self.norm_mha = nn.LayerNorm(size, eps=1e-12)
        self.self_attn = RelPositionMultiHeadedAttention(
            n_head, size, dropout_rate, key_bias
        )
        self.norm_ff = nn.LayerNorm(size, eps=1e-12)
        self.feed_forward = PositionwiseFeedForward(size, d_inner, dropout_rate)
        self.size = size

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        pos_emb: Optional[mx.array] = None,
    ) -> mx.array:
        """
        Args:
            x: Input (B, T, D)
            mask: Attention mask (B, T) or (B, T, T)
            pos_emb: Positional embedding (1, 2*T-1, D)

        Returns:
            Output (B, T, D)
        """
        # Multi-head self-attention with pre-norm and residual
        residual = x
        x = self.norm_mha(x)
        x = residual + self.self_attn(x, mask, pos_emb)

        # Feed-forward with pre-norm and residual
        residual = x
        x = self.norm_ff(x)
        x = residual + self.feed_forward(x)

        return x


class PreLookaheadLayer(nn.Module):
    """Pre-lookahead convolution layer."""

    def __init__(self, channels: int, pre_lookahead_len: int = 3):
        super().__init__()
        self.channels = channels
        self.pre_lookahead_len = pre_lookahead_len

        self.conv1 = nn.Conv1d(
            channels,
            channels,
            kernel_size=pre_lookahead_len + 1,
            stride=1,
            padding=0,
        )
        self.conv2 = nn.Conv1d(
            channels,
            channels,
            kernel_size=3,
            stride=1,
            padding=0,
        )

    def __call__(self, x: mx.array) -> mx.array:
        """
        Args:
            x: Input (B, T, C)

        Returns:
            Output (B, T, C)
        """
        # MLX Conv1d expects (B, T, C) format - no transpose needed
        # Look ahead padding on time dimension (axis 1)
        out = mx.pad(x, [(0, 0), (0, self.pre_lookahead_len), (0, 0)])
        out = nn.leaky_relu(self.conv1(out))

        # Causal padding for second conv
        out = mx.pad(out, [(0, 0), (2, 0), (0, 0)])
        out = self.conv2(out)

        # Residual connection
        return out + x


class Upsample1DEncoder(nn.Module):
    """1D upsampling layer for encoder."""

    def __init__(self, channels: int, stride: int = 2):
        super().__init__()
        self.channels = channels
        self.stride = stride
        self.conv = nn.Conv1d(
            channels, channels, kernel_size=stride * 2 + 1, stride=1, padding=0
        )

    def __call__(self, x: mx.array, x_lens: mx.array) -> Tuple[mx.array, mx.array]:
        """
        Args:
            x: Input (B, T, C) - MLX format
            x_lens: Lengths (B,)

        Returns:
            x: Upsampled (B, T*stride, C)
            x_lens: Updated lengths
        """
        B, T, C = x.shape

        # Nearest neighbor upsampling on time dimension
        x = mx.repeat(x, self.stride, axis=1)

        # Causal padding on time dimension and conv
        x = mx.pad(x, [(0, 0), (self.stride * 2, 0), (0, 0)])
        x = self.conv(x)

        return x, x_lens * self.stride


class UpsampleConformerEncoder(nn.Module):
    """
    Upsampling Conformer encoder for S3Gen.
    Converts speech tokens to mel-spectrogram features.
    """

    def __init__(
        self,
        input_size: int = 512,
        output_size: int = 512,
        attention_heads: int = 8,
        linear_units: int = 2048,
        num_blocks: int = 6,
        dropout_rate: float = 0.1,
    ):
        super().__init__()
        self._output_size = output_size

        # Input embedding
        self.embed = LinearInput(input_size, output_size, dropout_rate)

        # Pre-lookahead layer
        self.pre_lookahead_layer = PreLookaheadLayer(output_size, pre_lookahead_len=3)

        # Encoder layers
        self.encoders = [
            ConformerEncoderLayer(
                output_size, attention_heads, linear_units, dropout_rate
            )
            for _ in range(num_blocks)
        ]

        # Upsampling
        self.up_layer = Upsample1DEncoder(output_size, stride=2)

        # Post-upsample embedding
        self.up_embed = LinearInput(input_size, output_size, dropout_rate)

        # Post-upsample encoder layers
        self.up_encoders = [
            ConformerEncoderLayer(
                output_size, attention_heads, linear_units, dropout_rate
            )
            for _ in range(4)
        ]

        # Final norm
        self.after_norm = nn.LayerNorm(output_size)

    def output_size(self) -> int:
        return self._output_size

    def __call__(
        self,
        xs: mx.array,
        xs_lens: mx.array,
    ) -> Tuple[mx.array, mx.array]:
        """
        Args:
            xs: Input (B, T, D)
            xs_lens: Lengths (B,)

        Returns:
            xs: Encoded features (B, T*2, D)
            masks: Output masks (B, 1, T*2)
        """
        B, T, D = xs.shape

        # Create mask
        mask = mx.arange(T)[None, :] < xs_lens[:, None]  # (B, T)
        mask = mask[:, None, :]  # (B, 1, T)

        # Input projection
        xs, pos_emb, mask = self.embed(xs, mask)

        # Pre-lookahead
        xs = self.pre_lookahead_layer(xs)

        # Encoder layers
        mask_1d = mask[:, 0, :]  # (B, T)
        for layer in self.encoders:
            xs = layer(xs, mask_1d, pos_emb)

        # Upsampling - up_layer now works with (B, T, D) directly
        xs, xs_lens = self.up_layer(xs, xs_lens)

        # Update mask
        T2 = xs.shape[1]
        mask = mx.arange(T2)[None, :] < xs_lens[:, None]
        mask = mask[:, None, :]

        # Post-upsample embedding
        xs, pos_emb, mask = self.up_embed(xs, mask)

        # Post-upsample encoder
        mask_1d = mask[:, 0, :]
        for layer in self.up_encoders:
            xs = layer(xs, mask_1d, pos_emb)

        # Final norm
        xs = self.after_norm(xs)

        return xs, mask



import mlx.core as mx
import mlx.nn as nn


class GlobalLayerNorm(nn.Module):
    """MLX implementation of Global Layer Normalization.

    This class calculates Global Layer Normalization, providing identical
    behavior to the PyTorch version for both 3D and 4D tensors.

    Arguments:
        dim: Input dimension size
        shape: Number of dimensions (3 or 4)
        eps: Small value for numerical stability (default: 1e-8)
        elementwise_affine: Whether to use learnable affine parameters (default: True)
    """

    def __init__(
        self, dim: int, shape: int, eps: float = 1e-8, elementwise_affine: bool = True
    ):
        super().__init__()
        self.dim = dim
        self.shape = shape
        self.eps = eps
        self.elementwise_affine = elementwise_affine

        if self.elementwise_affine:
            if shape == 3:
                # Weight and bias for 3D tensors: initialize as [dim, 1] to match PyTorch
                # NPZ weights may be (dim,) or (dim, 1) - we'll handle both in loading
                self.weight = mx.ones((self.dim, 1))
                self.bias = mx.zeros((self.dim, 1))
            elif shape == 4:
                # Weight and bias for 4D tensors: initialize as [dim, 1, 1] to match PyTorch
                # NPZ weights may be (dim,) or (dim, 1) - we'll handle both in loading
                self.weight = mx.ones((self.dim, 1, 1))
                self.bias = mx.zeros((self.dim, 1, 1))
            else:
                raise ValueError(
                    f"Unsupported shape: {shape}. Only 3 and 4 are supported."
                )
        else:
            self.weight = None
            self.bias = None

    def __call__(self, x: mx.array) -> mx.array:
        """Forward pass for Global Layer Normalization.

        Args:
            x: Input tensor of size [N, C, L] for 3D or [N, C, K, S] for 4D

        Returns:
            Normalized tensor of the same shape as input
        """
        if x.ndim == 3:
            # For 3D tensors: normalize over dimensions (1, 2)
            # Input is [B, C, L] in PyTorch convention
            mean = mx.mean(x, axis=(1, 2), keepdims=True)
            var = mx.mean((x - mean) ** 2, axis=(1, 2), keepdims=True)

            if self.elementwise_affine:
                # Optimize broadcasting for weight and bias
                # Reshape weight and bias once for efficient broadcasting
                w = mx.reshape(
                    self.weight.squeeze() if self.weight.ndim > 1 else self.weight,
                    (1, -1, 1),
                )
                b = mx.reshape(
                    self.bias.squeeze() if self.bias.ndim > 1 else self.bias, (1, -1, 1)
                )
                # Fuse normalization and affine transformation
                x = w * (x - mean) / mx.sqrt(var + self.eps) + b
            else:
                x = (x - mean) / mx.sqrt(var + self.eps)

        elif x.ndim == 4:
            # For 4D tensors: normalize over dimensions (1, 2, 3)
            mean = mx.mean(x, axis=(1, 2, 3), keepdims=True)
            var = mx.mean((x - mean) ** 2, axis=(1, 2, 3), keepdims=True)

            if self.elementwise_affine:
                # Optimize broadcasting for weight and bias
                # Reshape weight and bias once for efficient broadcasting
                w = mx.reshape(
                    self.weight.squeeze() if self.weight.ndim > 1 else self.weight,
                    (1, -1, 1, 1),
                )
                b = mx.reshape(
                    self.bias.squeeze() if self.bias.ndim > 1 else self.bias,
                    (1, -1, 1, 1),
                )
                # Fuse normalization and affine transformation
                x = w * (x - mean) / mx.sqrt(var + self.eps) + b
            else:
                x = (x - mean) / mx.sqrt(var + self.eps)
        else:
            raise ValueError(f"Expected 3D or 4D tensor, got {x.ndim}D tensor")

        return x



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from dataclasses import dataclass
from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from .t3_config import T3Config


@dataclass
class T3Cond:
    """Dataclass container for conditioning info."""

    speaker_emb: mx.array
    clap_emb: Optional[mx.array] = None
    cond_prompt_speech_tokens: Optional[mx.array] = None
    cond_prompt_speech_emb: Optional[mx.array] = None
    emotion_adv: Optional[mx.array] = None


class T3CondEnc(nn.Module):
    """
    Handle all non-text conditioning, like speaker embeddings / prompts, CLAP, emotion, etc.
    """

    def __init__(self, hp: T3Config):
        super().__init__()
        self.hp = hp

        if hp.encoder_type == "voice_encoder":
            self.spkr_enc = nn.Linear(hp.speaker_embed_size, hp.n_channels)
        else:
            raise NotImplementedError(str(hp.encoder_type))

        # Emotion adv (not used in Turbo)
        self.emotion_adv_fc = None
        if hp.emotion_adv:
            self.emotion_adv_fc = nn.Linear(1, hp.n_channels, bias=False)

        # Perceiver resampler (not used in Turbo)
        self.perceiver = None

    def __call__(self, cond: T3Cond) -> mx.array:
        # Validate
        assert (cond.cond_prompt_speech_tokens is None) == (
            cond.cond_prompt_speech_emb is None
        ), "no embeddings for cond_prompt_speech_tokens"

        # Speaker embedding projection
        speaker_emb = cond.speaker_emb.reshape(-1, self.hp.speaker_embed_size)
        cond_spkr = self.spkr_enc(speaker_emb)[:, None, :]  # (B, 1, dim)

        B = cond_spkr.shape[0]
        dim = cond_spkr.shape[-1]

        # Empty tensor for unused conditions
        empty = mx.zeros((B, 0, dim))

        # CLAP (not implemented)
        assert cond.clap_emb is None, "clap_embed not implemented"
        cond_clap = empty

        # Cond prompt
        cond_prompt_speech_emb = cond.cond_prompt_speech_emb
        if cond_prompt_speech_emb is None:
            cond_prompt_speech_emb = empty
        elif self.hp.use_perceiver_resampler and self.perceiver is not None:
            cond_prompt_speech_emb = self.perceiver(cond_prompt_speech_emb)

        # Emotion Adv (not used in Turbo)
        cond_emotion_adv = empty
        if (
            self.hp.emotion_adv
            and cond.emotion_adv is not None
            and self.emotion_adv_fc is not None
        ):
            emotion_val = cond.emotion_adv.reshape(-1, 1, 1)
            cond_emotion_adv = self.emotion_adv_fc(emotion_val)

        # Concat and return
        cond_embeds = mx.concatenate(
            [
                cond_spkr,
                cond_clap,
                cond_prompt_speech_emb,
                cond_emotion_adv,
            ],
            axis=1,
        )

        return cond_embeds



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import unittest

import mlx.core as mx
import mlx.nn as nn

from mlx_audio.stt.models.vibevoice_asr.audio_encoder import (
    AcousticTokenizerEncoder,
    Block1D,
    ConvRMSNorm,
    SConv1d,
    SemanticTokenizerEncoder,
    TokenizerEncoder,
)
from mlx_audio.stt.models.vibevoice_asr.config import (
    AcousticTokenizerConfig,
    ModelConfig,
    Qwen2Config,
    SemanticTokenizerConfig,
)
from mlx_audio.stt.models.vibevoice_asr.vibevoice_asr import (
    LanguageModel,
    Model,
    SpeechConnector,
)


def _small_acoustic_config():
    """Create a small acoustic tokenizer config for fast testing."""
    return AcousticTokenizerConfig(
        channels=1,
        vae_dim=8,
        encoder_n_filters=4,
        encoder_ratios=[2, 2],
        encoder_depths="2-2-2",
        causal=True,
    )


def _small_semantic_config():
    """Create a small semantic tokenizer config for fast testing."""
    return SemanticTokenizerConfig(
        channels=1,
        vae_dim=16,
        encoder_n_filters=4,
        encoder_ratios=[2, 2],
        encoder_depths="2-2-2",
        causal=True,
    )


def _small_qwen2_config():
    """Create a small Qwen2 config for fast testing."""
    return Qwen2Config(
        vocab_size=1000,
        hidden_size=32,
        num_hidden_layers=2,
        num_attention_heads=4,
        num_key_value_heads=2,
        intermediate_size=64,
        tie_word_embeddings=False,
        layer_types=["full_attention"] * 2,
    )


def _small_model_config():
    """Create a small model config for fast testing."""
    return ModelConfig(
        acoustic_tokenizer_config=_small_acoustic_config(),
        semantic_tokenizer_config=_small_semantic_config(),
        decoder_config=_small_qwen2_config(),
        acoustic_vae_dim=8,
        semantic_vae_dim=16,
    )


class TestConfig(unittest.TestCase):
    """Tests for VibeVoice-ASR configuration classes."""

    def test_acoustic_tokenizer_config_from_dict(self):
        config_dict = {
            "channels": 1,
            "vae_dim": 64,
            "encoder_n_filters": 32,
            "encoder_ratios": [8, 5, 5, 4, 2, 2],
            "encoder_depths": "3-3-3-3-3-3-8",
            "causal": True,
            "fix_std": 0.5,
            "std_dist_type": "gaussian",
        }
        config = AcousticTokenizerConfig.from_dict(config_dict)

        self.assertEqual(config.channels, 1)
        self.assertEqual(config.vae_dim, 64)
        self.assertEqual(config.encoder_n_filters, 32)
        self.assertEqual(config.encoder_ratios, [8, 5, 5, 4, 2, 2])
        self.assertEqual(config.fix_std, 0.5)
        self.assertEqual(config.std_dist_type, "gaussian")

    def test_acoustic_tokenizer_parsed_depths(self):
        config = AcousticTokenizerConfig(encoder_depths="3-3-3-3-3-3-8")
        self.assertEqual(config.parsed_encoder_depths, [3, 3, 3, 3, 3, 3, 8])

    def test_acoustic_tokenizer_parsed_depths_list(self):
        config = AcousticTokenizerConfig(encoder_depths=[3, 3, 3])
        self.assertEqual(config.parsed_encoder_depths, [3, 3, 3])

    def test_semantic_tokenizer_config_from_dict(self):
        config_dict = {
            "channels": 1,
            "vae_dim": 128,
            "encoder_n_filters": 32,
            "encoder_ratios": [8, 5, 5, 4, 2, 2],
            "encoder_depths": "3-3-3-3-3-3-8",
            "std_dist_type": "none",
        }
        config = SemanticTokenizerConfig.from_dict(config_dict)

        self.assertEqual(config.vae_dim, 128)
        self.assertEqual(config.std_dist_type, "none")

    def test_qwen2_config_from_dict(self):
        config_dict = {
            "vocab_size": 152064,
            "hidden_size": 3584,
            "num_hidden_layers": 28,
            "num_attention_heads": 28,
            "num_key_value_heads": 4,
            "intermediate_size": 18944,
        }
        config = Qwen2Config.from_dict(config_dict)

        self.assertEqual(config.vocab_size, 152064)
        self.assertEqual(config.hidden_size, 3584)
        self.assertEqual(config.head_dim, 128)

    def test_qwen2_config_head_dim(self):
        config = Qwen2Config(hidden_size=256, num_attention_heads=8)
        self.assertEqual(config.head_dim, 32)

    def test_model_config_from_dict(self):
        config_dict = {
            "model_type": "vibevoice_asr",
            "acoustic_tokenizer_config": {
                "vae_dim": 64,
                "encoder_n_filters": 32,
            },
            "semantic_tokenizer_config": {
                "vae_dim": 128,
                "encoder_n_filters": 32,
            },
            "decoder_config": {
                "vocab_size": 152064,
                "hidden_size": 3584,
                "num_hidden_layers": 28,
                "num_attention_heads": 28,
                "num_key_value_heads": 4,
            },
            "acoustic_vae_dim": 64,
            "semantic_vae_dim": 128,
        }
        config = ModelConfig.from_dict(config_dict)

        self.assertEqual(config.model_type, "vibevoice_asr")
        self.assertIsInstance(config.acoustic_tokenizer_config, AcousticTokenizerConfig)
        self.assertIsInstance(config.semantic_tokenizer_config, SemanticTokenizerConfig)
        self.assertIsInstance(config.decoder_config, Qwen2Config)
        self.assertEqual(config.acoustic_tokenizer_config.vae_dim, 64)
        self.assertEqual(config.semantic_tokenizer_config.vae_dim, 128)
        self.assertEqual(config.decoder_config.hidden_size, 3584)

    def test_model_config_ignores_extra_keys(self):
        config_dict = {
            "model_type": "vibevoice_asr",
            "acoustic_tokenizer_config": {"vae_dim": 64, "unknown_key": 999},
            "semantic_tokenizer_config": {},
            "decoder_config": {"vocab_size": 1000, "extra_field": True},
        }
        config = ModelConfig.from_dict(config_dict)
        self.assertEqual(config.acoustic_tokenizer_config.vae_dim, 64)
        self.assertEqual(config.decoder_config.vocab_size, 1000)


class TestSConv1d(unittest.TestCase):
    """Tests for SConv1d (causal convolution)."""

    def test_output_shape_stride1(self):
        conv = SConv1d(in_channels=1, out_channels=32, kernel_size=7, stride=1)
        x = mx.random.normal((1, 100, 1))  # [B, T, C]
        y = conv(x)
        # Stride 1, causal padding preserves length
        self.assertEqual(y.shape[0], 1)
        self.assertEqual(y.shape[1], 100)
        self.assertEqual(y.shape[2], 32)

    def test_output_shape_stride2(self):
        conv = SConv1d(in_channels=4, out_channels=8, kernel_size=4, stride=2)
        x = mx.random.normal((1, 100, 4))
        y = conv(x)
        self.assertEqual(y.shape[0], 1)
        self.assertEqual(y.shape[1], 50)  # 100 / 2
        self.assertEqual(y.shape[2], 8)

    def test_groups_parameter(self):
        conv = SConv1d(in_channels=8, out_channels=8, kernel_size=7, stride=1, groups=8)
        x = mx.random.normal((1, 50, 8))
        y = conv(x)
        self.assertEqual(y.shape, (1, 50, 8))


class TestConvRMSNorm(unittest.TestCase):
    """Tests for ConvRMSNorm."""

    def test_output_shape(self):
        norm = ConvRMSNorm(dim=16)
        x = mx.random.normal((1, 16, 50))  # [B, C, T]
        y = norm(x)
        self.assertEqual(y.shape, x.shape)

    def test_normalization(self):
        norm = ConvRMSNorm(dim=4, elementwise_affine=False)
        x = mx.ones((1, 4, 10)) * 5.0
        y = norm(x)
        mx.eval(y)
        # RMSNorm of a constant should normalize it
        self.assertTrue(y.shape == x.shape)


class TestBlock1D(unittest.TestCase):
    """Tests for Block1D transformer block."""

    def test_output_shape(self):
        block = Block1D(dim=16, kernel_size=7, causal=True)
        x = mx.random.normal((1, 50, 16))  # [B, T, C]
        y = block(x)
        self.assertEqual(y.shape, x.shape)

    def test_residual_connection(self):
        block = Block1D(dim=8, kernel_size=3, causal=True, layer_scale_init_value=0.0)
        x = mx.zeros((1, 20, 8))
        y = block(x)
        mx.eval(y)
        # With zero input, output should be close to zero (residual + small FFN output)
        self.assertEqual(y.shape, x.shape)


class TestTokenizerEncoder(unittest.TestCase):
    """Tests for TokenizerEncoder."""

    def test_output_shape(self):
        encoder = TokenizerEncoder(
            channels=1,
            vae_dim=8,
            n_filters=4,
            ratios=[2, 2],
            depths=[2, 2, 2],
            causal=True,
        )
        # Input: [B, 1, T] - 1 channel, T samples
        x = mx.random.normal((1, 1, 160))
        y = encoder(x)
        mx.eval(y)

        # Total downsampling: product of ratios = 2*2 = 4
        # T' = 160 / 4 = 40
        self.assertEqual(y.shape[0], 1)  # batch
        self.assertEqual(y.shape[1], 40)  # time (downsampled)
        self.assertEqual(y.shape[2], 8)  # vae_dim

    def test_2d_input(self):
        encoder = TokenizerEncoder(
            channels=1, vae_dim=8, n_filters=4, ratios=[2], depths=[2, 2]
        )
        # [B, T] input without channel dim
        x = mx.random.normal((1, 80))
        y = encoder(x)
        mx.eval(y)

        self.assertEqual(y.shape[0], 1)
        self.assertEqual(y.shape[1], 40)  # 80 / 2
        self.assertEqual(y.shape[2], 8)

    def test_hop_length(self):
        encoder = TokenizerEncoder(
            channels=1, vae_dim=8, n_filters=4, ratios=[4, 2, 2], depths=[1, 1, 1, 1]
        )
        # hop_length should be product of ratios
        self.assertEqual(encoder.hop_length, 16)


class TestAcousticTokenizerEncoder(unittest.TestCase):
    """Tests for AcousticTokenizerEncoder."""

    def test_encode_output_shape(self):
        config = _small_acoustic_config()
        encoder = AcousticTokenizerEncoder(config)

        x = mx.random.normal((1, 1, 160))  # [B, 1, T]
        y = encoder.encode(x)
        mx.eval(y)

        # Downsampled by product of ratios (2*2=4): 160/4 = 40
        self.assertEqual(y.shape[0], 1)
        self.assertEqual(y.shape[1], 40)
        self.assertEqual(y.shape[2], config.vae_dim)

    def test_sample_gaussian(self):
        config = _small_acoustic_config()
        config.fix_std = 0.5
        config.std_dist_type = "gaussian"
        encoder = AcousticTokenizerEncoder(config)

        mean = mx.zeros((1, 10, config.vae_dim))
        sampled = encoder.sample(mean)
        mx.eval(sampled)

        self.assertEqual(sampled.shape, mean.shape)

    def test_sample_none(self):
        config = _small_acoustic_config()
        config.std_dist_type = "none"
        encoder = AcousticTokenizerEncoder(config)

        mean = mx.ones((1, 10, config.vae_dim))
        sampled = encoder.sample(mean)
        mx.eval(sampled)

        # With "none" dist type, output should equal input
        self.assertTrue(mx.allclose(sampled, mean).item())


class TestSemanticTokenizerEncoder(unittest.TestCase):
    """Tests for SemanticTokenizerEncoder."""

    def test_encode_output_shape(self):
        config = _small_semantic_config()
        encoder = SemanticTokenizerEncoder(config)

        x = mx.random.normal((1, 1, 160))
        y = encoder.encode(x)
        mx.eval(y)

        self.assertEqual(y.shape[0], 1)
        self.assertEqual(y.shape[1], 40)  # 160 / (2*2)
        self.assertEqual(y.shape[2], config.vae_dim)


class TestSpeechConnector(unittest.TestCase):
    """Tests for SpeechConnector (MLP projection)."""

    def test_output_shape(self):
        connector = SpeechConnector(input_dim=8, output_dim=32)
        x = mx.random.normal((1, 10, 8))
        y = connector(x)
        mx.eval(y)

        self.assertEqual(y.shape, (1, 10, 32))

    def test_different_dims(self):
        connector = SpeechConnector(input_dim=16, output_dim=64)
        x = mx.random.normal((2, 5, 16))
        y = connector(x)
        mx.eval(y)

        self.assertEqual(y.shape, (2, 5, 64))


class TestLanguageModel(unittest.TestCase):
    """Tests for LanguageModel (Qwen2 wrapper)."""

    def test_forward_pass(self):
        config = _small_qwen2_config()
        lm = LanguageModel(config)

        input_ids = mx.array([[1, 2, 3, 4, 5]])
        logits = lm(input_ids)
        mx.eval(logits)

        self.assertEqual(logits.shape, (1, 5, config.vocab_size))

    def test_forward_with_embeddings(self):
        config = _small_qwen2_config()
        lm = LanguageModel(config)

        embeddings = mx.random.normal((1, 5, config.hidden_size))
        logits = lm(input_embeddings=embeddings)
        mx.eval(logits)

        self.assertEqual(logits.shape, (1, 5, config.vocab_size))

    def test_embed_tokens_property(self):
        config = _small_qwen2_config()
        lm = LanguageModel(config)

        self.assertIsInstance(lm.embed_tokens, nn.Embedding)

    def test_layers_property(self):
        config = _small_qwen2_config()
        lm = LanguageModel(config)

        self.assertEqual(len(lm.layers), config.num_hidden_layers)


class TestModelSanitize(unittest.TestCase):
    """Tests for Model.sanitize weight processing."""

    def setUp(self):
        self.config = _small_model_config()
        self.model = Model(self.config)

    def test_removes_model_prefix(self):
        weights = {
            "model.acoustic_tokenizer.encoder.head.conv.weight": mx.zeros((8, 7, 16)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn("acoustic_tokenizer.encoder.head.conv.weight", sanitized)
        self.assertNotIn("model.acoustic_tokenizer.encoder.head.conv.weight", sanitized)

    def test_skips_decoder_weights(self):
        weights = {
            "model.acoustic_tokenizer.decoder.layers.0.weight": mx.zeros((10, 10)),
            "model.acoustic_tokenizer.encoder.head.conv.weight": mx.zeros((8, 7, 16)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertNotIn("acoustic_tokenizer.decoder.layers.0.weight", sanitized)
        self.assertIn("acoustic_tokenizer.encoder.head.conv.weight", sanitized)

    def test_downsample_layers_remap(self):
        weights = {
            "model.acoustic_tokenizer.encoder.downsample_layers.0.0.conv.conv.weight": mx.zeros(
                (4, 7, 1)
            ),
            "model.acoustic_tokenizer.encoder.downsample_layers.1.0.conv.conv.weight": mx.zeros(
                (8, 4, 4)
            ),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn(
            "acoustic_tokenizer.encoder.downsample_layers.0.conv.weight", sanitized
        )
        self.assertIn(
            "acoustic_tokenizer.encoder.downsample_layers.1.conv.weight", sanitized
        )

    def test_head_conv_remap(self):
        weights = {
            "model.acoustic_tokenizer.encoder.head.conv.conv.weight": mx.zeros(
                (8, 7, 16)
            ),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn("acoustic_tokenizer.encoder.head.conv.weight", sanitized)

    def test_mixer_conv_remap(self):
        weights = {
            "model.acoustic_tokenizer.encoder.stages.0.0.mixer.conv.conv.conv.weight": mx.zeros(
                (4, 7, 1)
            ),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn(
            "acoustic_tokenizer.encoder.stages.0.0.mixer.conv.conv.weight", sanitized
        )

    def test_language_model_layers_remap(self):
        weights = {
            "model.language_model.layers.0.self_attn.q_proj.weight": mx.zeros((32, 32)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn(
            "language_model.model.layers.0.self_attn.q_proj.weight", sanitized
        )

    def test_language_model_embed_tokens_remap(self):
        weights = {
            "model.language_model.embed_tokens.weight": mx.zeros((1000, 32)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn("language_model.model.embed_tokens.weight", sanitized)

    def test_language_model_norm_remap(self):
        weights = {
            "model.language_model.norm.weight": mx.zeros((32,)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn("language_model.model.norm.weight", sanitized)

    def test_lm_head_remap(self):
        weights = {
            "model.lm_head.weight": mx.zeros((1000, 32)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertIn("language_model.lm_head.weight", sanitized)

    def test_conv_transpose_pytorch_weights(self):
        """Test that PyTorch conv weights (with model. prefix) are transposed."""
        # PyTorch format: [out, in, kernel] -> MLX format: [out, kernel, in]
        pytorch_weight = mx.zeros((32, 1, 7))  # out=32, in=1, kernel=7
        weights = {
            "model.acoustic_tokenizer.encoder.downsample_layers.0.0.conv.conv.weight": pytorch_weight,
        }
        sanitized = self.model.sanitize(weights)

        key = "acoustic_tokenizer.encoder.downsample_layers.0.conv.weight"
        self.assertEqual(sanitized[key].shape, (32, 7, 1))  # Transposed

    def test_no_transpose_already_converted(self):
        """Test that already-converted MLX weights are NOT transposed."""
        # MLX format weight (no "model." prefix = already converted)
        mlx_weight = mx.zeros((32, 7, 1))  # out=32, kernel=7, in=1
        weights = {
            "acoustic_tokenizer.encoder.downsample_layers.0.conv.weight": mlx_weight,
        }
        sanitized = self.model.sanitize(weights)

        key = "acoustic_tokenizer.encoder.downsample_layers.0.conv.weight"
        self.assertEqual(sanitized[key].shape, (32, 7, 1))  # Unchanged

    def test_skips_position_ids(self):
        weights = {
            "model.language_model.layers.0.position_ids": mx.zeros((1, 128)),
            "model.language_model.norm.weight": mx.zeros((32,)),
        }
        sanitized = self.model.sanitize(weights)

        self.assertNotIn("language_model.model.layers.0.position_ids", sanitized)
        self.assertIn("language_model.model.norm.weight", sanitized)

    def test_skips_fix_std(self):
        weights = {
            "model.acoustic_tokenizer.fix_std": mx.array([0.5]),
        }
        sanitized = self.model.sanitize(weights)

        self.assertNotIn("acoustic_tokenizer.fix_std", sanitized)


class TestModelQuantPredicate(unittest.TestCase):
    """Tests for Model.model_quant_predicate."""

    def setUp(self):
        self.config = _small_model_config()
        self.model = Model(self.config)

    def test_quantize_language_model(self):
        self.assertTrue(
            self.model.model_quant_predicate(
                "language_model.model.layers.0.self_attn.q_proj", None
            )
        )

    def test_skip_acoustic_tokenizer(self):
        self.assertFalse(
            self.model.model_quant_predicate(
                "acoustic_tokenizer.encoder.downsample_layers.0.conv", None
            )
        )

    def test_skip_semantic_tokenizer(self):
        self.assertFalse(
            self.model.model_quant_predicate(
                "semantic_tokenizer.encoder.stages.0.0.ffn.linear1", None
            )
        )

    def test_skip_connectors(self):
        self.assertFalse(
            self.model.model_quant_predicate("acoustic_connector.fc1", None)
        )
        self.assertFalse(
            self.model.model_quant_predicate("semantic_connector.fc2", None)
        )


class TestModel(unittest.TestCase):
    """Tests for the full VibeVoice-ASR Model."""

    def setUp(self):
        self.config = _small_model_config()
        self.model = Model(self.config)

    def test_model_init(self):
        self.assertIsInstance(self.model.acoustic_tokenizer, AcousticTokenizerEncoder)
        self.assertIsInstance(self.model.semantic_tokenizer, SemanticTokenizerEncoder)
        self.assertIsInstance(self.model.acoustic_connector, SpeechConnector)
        self.assertIsInstance(self.model.semantic_connector, SpeechConnector)
        self.assertIsInstance(self.model.language_model, LanguageModel)

    def test_get_input_embeddings(self):
        emb = self.model.get_input_embeddings()
        self.assertIsInstance(emb, nn.Embedding)

    def test_encode_speech_1d(self):
        """Test encode_speech with 1D input [T]."""
        # Small audio: 160 samples
        audio = mx.random.normal((160,))
        features = self.model.encode_speech(audio)
        mx.eval(features)

        # Output should be [1, T', hidden_size]
        self.assertEqual(features.shape[0], 1)
        self.assertEqual(features.shape[2], self.config.decoder_config.hidden_size)

    def test_encode_speech_2d(self):
        """Test encode_speech with 2D input [B, T]."""
        audio = mx.random.normal((1, 160))
        features = self.model.encode_speech(audio)
        mx.eval(features)

        self.assertEqual(features.shape[0], 1)
        self.assertEqual(features.shape[2], self.config.decoder_config.hidden_size)

    def test_encode_speech_3d(self):
        """Test encode_speech with 3D input [B, 1, T]."""
        audio = mx.random.normal((1, 1, 160))
        features = self.model.encode_speech(audio)
        mx.eval(features)

        self.assertEqual(features.shape[0], 1)
        self.assertEqual(features.shape[2], self.config.decoder_config.hidden_size)

    def test_language_model_forward(self):
        """Test that language model produces correct logit shape."""
        input_ids = mx.array([[1, 2, 3, 4, 5]])
        logits = self.model.language_model(input_ids)
        mx.eval(logits)

        self.assertEqual(logits.shape, (1, 5, self.config.decoder_config.vocab_size))

    def test_merge_speech_text_embeddings(self):
        """Test embedding merging with speech features."""
        seq_len = 10
        speech_len = 4
        hidden = self.config.decoder_config.hidden_size

        input_ids = mx.ones((1, seq_len), dtype=mx.int32)
        speech_features = mx.random.normal((1, speech_len, hidden))
        # Mask: positions 2-5 are speech
        mask = mx.array(
            [[False, False, True, True, True, True, False, False, False, False]]
        )

        embeddings = self.model._merge_speech_text_embeddings(
            input_ids, speech_features, mask
        )
        mx.eval(embeddings)

        self.assertEqual(embeddings.shape, (1, seq_len, hidden))

    def test_merge_no_speech_features(self):
        """Test embedding merging without speech features returns text embeddings."""
        input_ids = mx.array([[1, 2, 3]])
        embeddings = self.model._merge_speech_text_embeddings(input_ids, None, None)
        mx.eval(embeddings)

        self.assertEqual(
            embeddings.shape, (1, 3, self.config.decoder_config.hidden_size)
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025 Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from dataclasses import dataclass, field
from typing import Optional

from mlx_audio.codec.models.dacvae.codec import DACVAEConfig


@dataclass
class T5EncoderConfig:
    """Configuration for the T5 text encoder."""

    name: str = "t5-base"
    max_length: Optional[int] = 512
    pad_mode: str = "longest"
    dim: int = 768


@dataclass
class TransformerConfig:
    """Configuration for the DiT (Diffusion Transformer).

    Note: SAM-Audio operates in codebook space (128 dim), so:
    - out_channels = 2 * codebook_dim = 256 (target + residual)
    - context_dim = transformer.dim (memory is projected to this dim)
    """

    dim: int = 2816  # Default from sam-audio-large
    n_heads: int = 22
    n_layers: int = 22
    dropout: float = 0.1
    norm_eps: float = 1.0e-05
    qk_norm: bool = True
    fc_bias: bool = False
    ffn_exp: int = 4
    ffn_dim_multiplier: int = 1
    multiple_of: int = 64
    non_linearity: str = "swiglu"
    use_rope: bool = True
    max_positions: int = 10000
    frequency_embedding_dim: int = 256
    timestep_non_linearity: str = "swiglu"
    t_block_non_linearity: str = "silu"
    t_block_bias: bool = True
    # context_dim should match transformer dim (memory is projected before passing to DiT)
    context_dim: int = 2816
    context_non_linearity: str = "swiglu"
    context_embedder_dropout: float = 0.0
    context_norm: bool = False
    # out_channels = 2 * codebook_dim (for target + residual features in codebook space)
    out_channels: int = 256
    in_channels: Optional[int] = None


@dataclass
class SAMAudioConfig:
    """Main configuration for SAMAudio model.

    Note: SAM-Audio operates in the VAE codebook space (codebook_dim=128), not the
    raw latent space (latent_dim=1024). This means:
    - in_channels = 6 * codebook_dim = 768 (for [noisy, zeros, features] x 2)
    - out_channels = 2 * codebook_dim = 256 (for target + residual)
    """

    # in_channels = 6 * codebook_dim (for concat of [noisy, zeros, features] where features is doubled)
    # Default: 6 * 128 = 768
    in_channels: int = 768
    audio_codec: DACVAEConfig = field(default_factory=DACVAEConfig)
    text_encoder: T5EncoderConfig = field(default_factory=T5EncoderConfig)
    transformer: TransformerConfig = field(default_factory=TransformerConfig)
    num_anchors: int = 3
    anchor_embedding_dim: int = 128

    @classmethod
    def from_dict(cls, config_dict: dict) -> "SAMAudioConfig":
        """Create config from dictionary."""
        audio_codec = DACVAEConfig(**config_dict.get("audio_codec", {}))
        text_encoder = T5EncoderConfig(**config_dict.get("text_encoder", {}))

        # Get transformer config - use values from config as-is
        transformer_dict = config_dict.get("transformer", {}).copy()
        transformer = TransformerConfig(**transformer_dict)

        # Use in_channels from config
        # SAM-Audio operates in codebook space: in_channels = 6 * codebook_dim = 768
        in_channels = config_dict.get("in_channels", 6 * audio_codec.codebook_dim)

        return cls(
            in_channels=in_channels,
            audio_codec=audio_codec,
            text_encoder=text_encoder,
            transformer=transformer,
            num_anchors=config_dict.get("num_anchors", 3),
            anchor_embedding_dim=config_dict.get("anchor_embedding_dim", 128),
        )


# Predefined configurations for different model sizes
# Note: SAM-Audio operates in codebook space (128 dim)
# - in_channels = 6 * codebook_dim = 768
# - out_channels = 2 * codebook_dim = 256
# - context_dim should match transformer dim (memory is projected before DiT)
SAM_AUDIO_SMALL_CONFIG = SAMAudioConfig(
    in_channels=768,  # 6 * codebook_dim
    audio_codec=DACVAEConfig(),
    text_encoder=T5EncoderConfig(name="t5-base", dim=768),
    transformer=TransformerConfig(
        dim=1024,
        n_heads=8,
        n_layers=12,
        context_dim=1024,  # Match transformer dim
        out_channels=256,  # 2 * codebook_dim
    ),
)

SAM_AUDIO_BASE_CONFIG = SAMAudioConfig(
    in_channels=768,  # 6 * codebook_dim
    audio_codec=DACVAEConfig(),
    text_encoder=T5EncoderConfig(name="t5-base", dim=768),
    transformer=TransformerConfig(
        dim=1536,
        n_heads=12,
        n_layers=16,
        context_dim=1536,  # Match transformer dim
        out_channels=256,  # 2 * codebook_dim
    ),
)

SAM_AUDIO_LARGE_CONFIG = SAMAudioConfig(
    in_channels=768,  # 6 * codebook_dim
    audio_codec=DACVAEConfig(),
    text_encoder=T5EncoderConfig(name="t5-base", dim=768),
    transformer=TransformerConfig(
        dim=2816,
        n_heads=22,
        n_layers=22,
        context_dim=2816,  # Match transformer dim
        out_channels=256,  # 2 * codebook_dim
    ),
)



# Copyright 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import unicodedata
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from .qwen3_asr import AudioEncoder, TextModel, _get_feat_extract_output_lengths


class ForceAlignProcessor:
    """Processor for forced alignment text tokenization and timestamp parsing."""

    def __init__(self):
        self.ko_score = None
        self.ko_tokenizer = None
        self._nagisa = None

    def is_kept_char(self, ch: str) -> bool:
        """Check if character should be kept (letters, numbers, apostrophe)."""
        if ch == "'":
            return True
        cat = unicodedata.category(ch)
        if cat.startswith("L") or cat.startswith("N"):
            return True
        return False

    def clean_token(self, token: str) -> str:
        """Remove non-kept characters from token."""
        return "".join(ch for ch in token if self.is_kept_char(ch))

    def is_cjk_char(self, ch: str) -> bool:
        """Check if character is CJK (Chinese/Japanese/Korean ideograph)."""
        code = ord(ch)
        return (
            0x4E00 <= code <= 0x9FFF  # CJK Unified Ideographs
            or 0x3400 <= code <= 0x4DBF  # Extension A
            or 0x20000 <= code <= 0x2A6DF  # Extension B
            or 0x2A700 <= code <= 0x2B73F  # Extension C
            or 0x2B740 <= code <= 0x2B81F  # Extension D
            or 0x2B820 <= code <= 0x2CEAF  # Extension E
            or 0xF900 <= code <= 0xFAFF  # Compatibility Ideographs
        )

    def tokenize_chinese_mixed(self, text: str) -> List[str]:
        """Tokenize text with Chinese characters (each char is a token)."""
        tokens: List[str] = []
        current_latin: List[str] = []

        def flush_latin():
            nonlocal current_latin
            if current_latin:
                token = "".join(current_latin)
                cleaned = self.clean_token(token)
                if cleaned:
                    tokens.append(cleaned)
                current_latin = []

        for ch in text:
            if self.is_cjk_char(ch):
                flush_latin()
                tokens.append(ch)
            else:
                if self.is_kept_char(ch):
                    current_latin.append(ch)
                else:
                    flush_latin()

        flush_latin()
        return tokens

    def tokenize_japanese(self, text: str) -> List[str]:
        """Tokenize Japanese text using nagisa."""
        if self._nagisa is None:
            try:
                import nagisa

                self._nagisa = nagisa
            except ImportError:
                raise ImportError(
                    "Japanese tokenization requires nagisa. "
                    "Install with: pip install nagisa"
                )
        words = self._nagisa.tagging(text).words
        tokens: List[str] = []
        for w in words:
            cleaned = self.clean_token(w)
            if cleaned:
                tokens.append(cleaned)
        return tokens

    def tokenize_korean(self, text: str) -> List[str]:
        """Tokenize Korean text using soynlp."""
        if self.ko_tokenizer is None:
            try:
                from soynlp.tokenizer import LTokenizer

                # Simple frequency-based tokenizer
                self.ko_tokenizer = LTokenizer()
            except ImportError:
                raise ImportError(
                    "Korean tokenization requires soynlp. "
                    "Install with: pip install soynlp"
                )
        raw_tokens = self.ko_tokenizer.tokenize(text)
        tokens: List[str] = []
        for w in raw_tokens:
            w_clean = self.clean_token(w)
            if w_clean:
                tokens.append(w_clean)
        return tokens

    def split_segment_with_chinese(self, seg: str) -> List[str]:
        """Split segment containing Chinese characters."""
        tokens: List[str] = []
        buf: List[str] = []

        def flush_buf():
            nonlocal buf
            if buf:
                tokens.append("".join(buf))
                buf = []

        for ch in seg:
            if self.is_cjk_char(ch):
                flush_buf()
                tokens.append(ch)
            else:
                buf.append(ch)

        flush_buf()
        return tokens

    def tokenize_space_lang(self, text: str) -> List[str]:
        """Tokenize space-separated languages (English, etc.)."""
        tokens: List[str] = []
        for seg in text.split():
            cleaned = self.clean_token(seg)
            if cleaned:
                tokens.extend(self.split_segment_with_chinese(cleaned))
        return tokens

    def fix_timestamp(self, data: np.ndarray) -> List[int]:
        """Fix non-monotonic timestamps using Longest Increasing Subsequence."""
        data = data.tolist()
        n = len(data)

        if n == 0:
            return []

        # Find LIS using dynamic programming
        dp = [1] * n
        parent = [-1] * n

        for i in range(1, n):
            for j in range(i):
                if data[j] <= data[i] and dp[j] + 1 > dp[i]:
                    dp[i] = dp[j] + 1
                    parent[i] = j

        max_length = max(dp)
        max_idx = dp.index(max_length)

        # Reconstruct LIS indices
        lis_indices = []
        idx = max_idx
        while idx != -1:
            lis_indices.append(idx)
            idx = parent[idx]
        lis_indices.reverse()

        is_normal = [False] * n
        for idx in lis_indices:
            is_normal[idx] = True

        result = data.copy()
        i = 0

        while i < n:
            if not is_normal[i]:
                j = i
                while j < n and not is_normal[j]:
                    j += 1

                anomaly_count = j - i

                if anomaly_count <= 2:
                    # For small anomalies, use nearest valid neighbor
                    left_val = None
                    for k in range(i - 1, -1, -1):
                        if is_normal[k]:
                            left_val = result[k]
                            break

                    right_val = None
                    for k in range(j, n):
                        if is_normal[k]:
                            right_val = result[k]
                            break

                    for k in range(i, j):
                        if left_val is None:
                            result[k] = right_val
                        elif right_val is None:
                            result[k] = left_val
                        else:
                            result[k] = (
                                left_val if (k - (i - 1)) <= ((j) - k) else right_val
                            )

                else:
                    # For large anomalies, interpolate linearly
                    left_val = None
                    for k in range(i - 1, -1, -1):
                        if is_normal[k]:
                            left_val = result[k]
                            break

                    right_val = None
                    for k in range(j, n):
                        if is_normal[k]:
                            right_val = result[k]
                            break

                    if left_val is not None and right_val is not None:
                        step = (right_val - left_val) / (anomaly_count + 1)
                        for k in range(i, j):
                            result[k] = left_val + step * (k - i + 1)
                    elif left_val is not None:
                        for k in range(i, j):
                            result[k] = left_val
                    elif right_val is not None:
                        for k in range(i, j):
                            result[k] = right_val

                i = j
            else:
                i += 1

        return [int(res) for res in result]

    def encode_timestamp(self, text: str, language: str) -> Tuple[List[str], str]:
        """Tokenize text and create input with timestamp tokens.

        Args:
            text: The transcript text to align.
            language: Language name (e.g., "Chinese", "Japanese", "Korean", "English").

        Returns:
            Tuple of (word_list, input_text with timestamp tokens).
        """
        language = language.lower()

        if language == "japanese":
            word_list = self.tokenize_japanese(text)
        elif language == "korean":
            word_list = self.tokenize_korean(text)
        elif language == "chinese":
            word_list = self.tokenize_chinese_mixed(text)
        else:
            # Space-separated languages (English, etc.)
            word_list = self.tokenize_space_lang(text)

        # Build input text with timestamp tokens between words
        input_text = "<timestamp><timestamp>".join(word_list) + "<timestamp><timestamp>"
        input_text = "<|audio_start|><|audio_pad|><|audio_end|>" + input_text

        return word_list, input_text

    def parse_timestamp(
        self, word_list: List[str], timestamp: np.ndarray
    ) -> List[Dict[str, Any]]:
        """Parse timestamps into word-level alignments.

        Args:
            word_list: List of words/tokens.
            timestamp: Raw timestamp values in milliseconds.

        Returns:
            List of dicts with text, start_time, end_time.
        """
        timestamp_output = []
        timestamp_fixed = self.fix_timestamp(timestamp)

        for i, word in enumerate(word_list):
            start_time = timestamp_fixed[i * 2]
            end_time = timestamp_fixed[i * 2 + 1]
            timestamp_output.append(
                {"text": word, "start_time": start_time, "end_time": end_time}
            )

        return timestamp_output


@dataclass(frozen=True)
class ForcedAlignItem:
    """One aligned item span.

    Attributes:
        text: The aligned unit (CJK character or word).
        start_time: Start time in seconds.
        end_time: End time in seconds.
    """

    text: str
    start_time: float
    end_time: float


@dataclass(frozen=True)
class ForcedAlignResult:
    """Forced alignment output for one sample.

    Attributes:
        items: Aligned token spans.
    """

    items: List[ForcedAlignItem]

    @property
    def text(self) -> str:
        """Full text from all aligned items."""
        return " ".join(item.text for item in self.items)

    @property
    def segments(self) -> List[Dict[str, Any]]:
        """Segments in STTOutput-compatible format."""
        return [
            {"text": item.text, "start": item.start_time, "end": item.end_time}
            for item in self.items
        ]

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, idx: int) -> ForcedAlignItem:
        return self.items[idx]


@dataclass
class ForcedAlignerConfig:
    """Configuration for Qwen3-ForcedAligner model."""

    audio_config: Any = None
    text_config: Any = None
    model_type: str = "qwen3_forced_aligner"
    model_repo: str = None
    audio_token_id: int = 151676
    audio_start_token_id: int = 151669
    audio_end_token_id: int = 151670
    timestamp_token_id: int = 151705
    timestamp_segment_time: float = 80.0
    classify_num: int = 5000  # Number of timestamp classes
    support_languages: List[str] = None

    def __post_init__(self):
        from .config import AudioEncoderConfig, TextConfig

        if self.audio_config is None:
            self.audio_config = AudioEncoderConfig()
        elif isinstance(self.audio_config, dict):
            self.audio_config = AudioEncoderConfig.from_dict(self.audio_config)

        if self.text_config is None:
            self.text_config = TextConfig()
        elif isinstance(self.text_config, dict):
            self.text_config = TextConfig.from_dict(self.text_config)

        if self.support_languages is None:
            self.support_languages = []

    @classmethod
    def from_dict(cls, params: Dict[str, Any]) -> "ForcedAlignerConfig":
        from .config import AudioEncoderConfig, TextConfig

        params = params.copy()

        # Extract top-level timestamp params (HF config has these at root level)
        top_level_timestamp_token_id = params.get("timestamp_token_id")
        top_level_timestamp_segment_time = params.get("timestamp_segment_time")

        # Handle nested thinker_config (from HF config)
        if "thinker_config" in params:
            thinker = params.pop("thinker_config")
            if "audio_config" in thinker:
                params["audio_config"] = thinker["audio_config"]
            if "text_config" in thinker:
                params["text_config"] = thinker["text_config"]
            if "audio_token_id" in thinker:
                params["audio_token_id"] = thinker["audio_token_id"]
            if "audio_start_token_id" in thinker:
                params["audio_start_token_id"] = thinker["audio_start_token_id"]
            if "audio_end_token_id" in thinker:
                params["audio_end_token_id"] = thinker["audio_end_token_id"]
            # Use thinker values if present, otherwise use top-level
            if "timestamp_token_id" in thinker:
                params["timestamp_token_id"] = thinker["timestamp_token_id"]
            if "timestamp_segment_time" in thinker:
                params["timestamp_segment_time"] = thinker["timestamp_segment_time"]
            if "classify_num" in thinker:
                params["classify_num"] = thinker["classify_num"]

        # Use top-level timestamp params if not set from thinker_config
        if (
            top_level_timestamp_token_id is not None
            and "timestamp_token_id" not in params
        ):
            params["timestamp_token_id"] = top_level_timestamp_token_id
        if (
            top_level_timestamp_segment_time is not None
            and "timestamp_segment_time" not in params
        ):
            params["timestamp_segment_time"] = top_level_timestamp_segment_time

        # Handle nested configs
        if "audio_config" in params and isinstance(params["audio_config"], dict):
            params["audio_config"] = AudioEncoderConfig.from_dict(
                params["audio_config"]
            )
        elif "audio_config" not in params:
            params["audio_config"] = AudioEncoderConfig()

        if "text_config" in params and isinstance(params["text_config"], dict):
            params["text_config"] = TextConfig.from_dict(params["text_config"])
        elif "text_config" not in params:
            params["text_config"] = TextConfig()

        import inspect

        # Always set model_type to qwen3_forced_aligner for this config class
        params["model_type"] = "qwen3_forced_aligner"

        return cls(
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            }
        )


class ForcedAlignerModel(nn.Module):
    """Qwen3-ForcedAligner Model for word-level audio alignment.

    This model takes audio and text as input and returns word-level
    timestamps showing when each word/character is spoken.
    """

    def __init__(self, config: ForcedAlignerConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.text_config.vocab_size
        self.audio_tower = AudioEncoder(config.audio_config)
        self.model = TextModel(config.text_config)
        self.aligner_processor = ForceAlignProcessor()

        # ForcedAligner always uses lm_head for timestamp classification
        # The output size is classify_num (number of timestamp classes), not vocab_size
        self.lm_head = nn.Linear(
            config.text_config.hidden_size,
            config.classify_num,
            bias=False,
        )

    def get_audio_features(
        self,
        input_features: mx.array,
        feature_attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        """Encode audio features."""
        return self.audio_tower(input_features, feature_attention_mask)

    def __call__(
        self,
        input_ids: mx.array,
        input_features: Optional[mx.array] = None,
        feature_attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        """Forward pass returning logits.

        Args:
            input_ids: Token IDs [batch, seq_len].
            input_features: Audio features [batch, mel_bins, time].
            feature_attention_mask: Attention mask for audio.

        Returns:
            Logits tensor [batch, seq_len, classify_num] for timestamp classification.
        """
        inputs_embeds = self.model.embed_tokens(input_ids)

        if input_features is not None:
            audio_features = self.get_audio_features(
                input_features, feature_attention_mask
            )
            audio_features = audio_features.astype(inputs_embeds.dtype)

            audio_token_mask = input_ids == self.config.audio_token_id

            if audio_token_mask.any():
                batch_size, seq_len, hidden_dim = inputs_embeds.shape
                flat_mask = audio_token_mask.flatten()
                flat_mask_np = np.array(flat_mask)
                audio_indices = np.where(flat_mask_np)[0]

                if len(audio_indices) > 0 and audio_features.shape[0] > 0:
                    num_to_replace = min(len(audio_indices), audio_features.shape[0])
                    flat_embeds = inputs_embeds.reshape(-1, hidden_dim)

                    result_list = []
                    audio_idx = 0
                    for i in range(flat_embeds.shape[0]):
                        if audio_idx < num_to_replace and i == audio_indices[audio_idx]:
                            result_list.append(audio_features[audio_idx])
                            audio_idx += 1
                        else:
                            result_list.append(flat_embeds[i])

                    inputs_embeds = mx.stack(result_list, axis=0).reshape(
                        batch_size, seq_len, hidden_dim
                    )

        hidden_states = self.model(inputs_embeds=inputs_embeds, cache=None)

        # ForcedAligner always uses lm_head for timestamp classification
        logits = self.lm_head(hidden_states)

        return logits

    @property
    def layers(self):
        return self.model.layers

    @property
    def sample_rate(self) -> int:
        """Sample rate for audio input."""
        return 16000

    @staticmethod
    def sanitize(weights: Dict[str, mx.array]) -> Dict[str, mx.array]:
        """Sanitize weights from HuggingFace/PyTorch format to MLX format."""
        sanitized = {}
        is_formatted = not any(k.startswith("thinker.") for k in weights.keys())

        for k, v in weights.items():
            if k.startswith("thinker."):
                k = k[len("thinker.") :]

            # ForcedAligner uses lm_head, don't skip it

            if (
                not is_formatted
                and "conv2d" in k
                and "weight" in k
                and len(v.shape) == 4
            ):
                v = v.transpose(0, 2, 3, 1)

            sanitized[k] = v

        return sanitized

    def model_quant_predicate(self, p: str, m: nn.Module) -> bool:
        """Determine which layers to quantize."""
        return not p.startswith("audio_tower")

    @classmethod
    def post_load_hook(
        cls, model: "ForcedAlignerModel", model_path: Path
    ) -> "ForcedAlignerModel":
        """Hook called after model weights are loaded."""
        import transformers
        from transformers import AutoTokenizer, WhisperFeatureExtractor

        # Suppress the harmless warning about model_type mismatch when loading
        # tokenizer for custom model types not registered in transformers
        prev_verbosity = transformers.logging.get_verbosity()
        transformers.logging.set_verbosity_error()
        try:
            model._tokenizer = AutoTokenizer.from_pretrained(
                str(model_path), trust_remote_code=True
            )
            model._feature_extractor = WhisperFeatureExtractor.from_pretrained(
                str(model_path)
            )
        finally:
            transformers.logging.set_verbosity(prev_verbosity)

        if not hasattr(model.config, "model_repo") or model.config.model_repo is None:
            try:
                index = model_path.parts.index("hub")
                model.config.model_repo = (
                    model_path.parts[index + 1]
                    .replace("models--", "")
                    .replace("--", "/")
                )
            except (ValueError, IndexError):
                model.config.model_repo = str(model_path)

        return model

    def _preprocess_audio(
        self,
        audio: Union[str, mx.array, np.ndarray],
    ) -> Tuple[mx.array, mx.array, int]:
        """Preprocess audio for the model."""
        from mlx_audio.stt.utils import load_audio

        if isinstance(audio, str):
            audio = load_audio(audio)

        audio_np = np.array(audio) if isinstance(audio, mx.array) else audio

        audio_inputs = self._feature_extractor(
            audio_np,
            sampling_rate=16000,
            return_attention_mask=True,
            truncation=False,
            padding=True,
            return_tensors="np",
        )
        input_features = mx.array(audio_inputs["input_features"])
        feature_attention_mask = mx.array(audio_inputs["attention_mask"])

        audio_lengths = feature_attention_mask.sum(axis=-1)
        aftercnn_lens = _get_feat_extract_output_lengths(audio_lengths)
        num_audio_tokens = int(aftercnn_lens[0].item())

        return input_features, feature_attention_mask, num_audio_tokens

    def generate(
        self,
        audio: Union[str, mx.array, np.ndarray, List[Union[str, mx.array, np.ndarray]]],
        text: Union[str, List[str]],
        language: Union[str, List[str]] = "English",
        **kwargs,
    ) -> Union[ForcedAlignResult, List[ForcedAlignResult]]:
        """Run forced alignment for audio and text.

        Args:
            audio: Audio input(s) - file path, array, or list of these.
            text: Transcript(s) for alignment.
            language: Language(s) for each sample (e.g., "Chinese", "English", "Japanese", "Korean").
            **kwargs: Additional arguments (ignored, for API compatibility).

        Returns:
            ForcedAlignResult or list of results with word-level timestamps.
        """
        from mlx_audio.stt.utils import load_audio

        if not hasattr(self, "_tokenizer") or not hasattr(self, "_feature_extractor"):
            raise RuntimeError(
                "Tokenizer/FeatureExtractor not initialized. Call post_load_hook first."
            )

        # Normalize inputs to lists
        single_input = not isinstance(audio, list)
        audios = [audio] if single_input else audio
        texts = [text] if isinstance(text, str) else text
        languages = [language] if isinstance(language, str) else language

        if len(languages) == 1 and len(audios) > 1:
            languages = languages * len(audios)

        if not (len(audios) == len(texts) == len(languages)):
            raise ValueError(
                f"Batch size mismatch: audio={len(audios)}, text={len(texts)}, language={len(languages)}"
            )

        results: List[ForcedAlignResult] = []

        # Process each sample
        for audio_input, txt, lang in zip(audios, texts, languages):
            # Load audio if needed
            if isinstance(audio_input, str):
                audio_input = load_audio(audio_input)
            audio_np = (
                np.array(audio_input)
                if isinstance(audio_input, mx.array)
                else audio_input
            )

            # Preprocess audio
            input_features, feature_attention_mask, num_audio_tokens = (
                self._preprocess_audio(audio_np)
            )

            # Encode text with timestamp tokens
            word_list, aligner_input_text = self.aligner_processor.encode_timestamp(
                txt, lang
            )

            # Replace single audio_pad with correct number
            aligner_input_text = aligner_input_text.replace(
                "<|audio_pad|>", "<|audio_pad|>" * num_audio_tokens
            )

            # Tokenize input
            input_ids = self._tokenizer.encode(
                aligner_input_text, return_tensors="np", add_special_tokens=False
            )
            input_ids = mx.array(input_ids)

            # Forward pass
            logits = self(
                input_ids,
                input_features=input_features,
                feature_attention_mask=feature_attention_mask,
            )
            mx.eval(logits)

            # Get predicted tokens
            output_ids = mx.argmax(logits, axis=-1)

            # Extract timestamps at timestamp token positions
            timestamp_token_id = self.config.timestamp_token_id
            timestamp_segment_time = self.config.timestamp_segment_time

            input_ids_flat = input_ids[0] if input_ids.ndim > 1 else input_ids
            output_ids_flat = output_ids[0] if output_ids.ndim > 1 else output_ids

            # Convert to numpy for boolean indexing (not supported in MLX)
            input_ids_np = np.array(input_ids_flat)
            output_ids_np = np.array(output_ids_flat)

            # Find positions where input has timestamp token
            timestamp_mask = input_ids_np == timestamp_token_id
            masked_output = output_ids_np[timestamp_mask]

            # Convert to milliseconds
            timestamp_ms = masked_output * timestamp_segment_time

            # Parse timestamps
            timestamp_output = self.aligner_processor.parse_timestamp(
                word_list, timestamp_ms
            )

            # Convert to seconds
            items = []
            for it in timestamp_output:
                items.append(
                    ForcedAlignItem(
                        text=str(it["text"]),
                        start_time=round(it["start_time"] / 1000.0, 3),
                        end_time=round(it["end_time"] / 1000.0, 3),
                    )
                )

            results.append(ForcedAlignResult(items=items))

            # Clear cache between samples
            mx.clear_cache()

        return results[0] if single_input else results

    def get_supported_languages(self) -> Optional[List[str]]:
        """List supported language names for the current model.

        Returns:
            Sorted list of supported languages, or None.
        """
        if hasattr(self.config, "support_languages") and self.config.support_languages:
            return sorted({str(x).lower() for x in self.config.support_languages})
        return None


# Alias for model loading compatibility
Model = ForcedAlignerModel



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from .decoder import ConditionalDecoder
from .encoder import UpsampleConformerEncoder
from .flow_matching import CausalConditionalCFM
from .hifigan import HiFTGenerator
from .mel import mel_spectrogram
from .s3gen import (
    S3GEN_SIL,
    S3GEN_SR,
    SPEECH_VOCAB_SIZE,
    S3Gen,
    S3Token2Mel,
    S3Token2Wav,
)

__all__ = [
    "S3Gen",
    "S3Token2Mel",
    "S3Token2Wav",
    "S3GEN_SR",
    "S3GEN_SIL",
    "SPEECH_VOCAB_SIZE",
    "HiFTGenerator",
    "UpsampleConformerEncoder",
    "ConditionalDecoder",
    "CausalConditionalCFM",
    "mel_spectrogram",
]



# Copyright (c) 2025 Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)
# LFM2.5-Audio: FastConformer encoder implementation

import math
from typing import List, Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from .config import ConformerEncoderConfig


class RelativePositionalEncoding(nn.Module):

    def __init__(self, d_model: int, max_len: int = 5000, xscale: bool = True):
        super().__init__()
        self.d_model = d_model
        self.max_len = max_len
        self.xscale = math.sqrt(d_model) if xscale else None
        self._pe_cache_len = 0
        self.pe = None

        # div_term for sinusoidal encoding: 10000^(-2i/d_model)
        self._div_term = mx.exp(
            mx.arange(0, d_model, 2, dtype=mx.float32) * (-math.log(10000.0) / d_model)
        )

    def _extend_pe(self, length: int):
        """Extend positional encodings if needed."""
        needed_size = 2 * length - 1
        if self.pe is not None and self.pe.shape[0] >= needed_size:
            return

        # Generate positions from (length-1) to -(length-1) in descending order
        # E.g., for length=39: [38, 37, ..., 1, 0, -1, ..., -37, -38]
        positions = mx.arange(length - 1, -length, -1, dtype=mx.float32)[:, None]

        # Create positional encoding
        pe = mx.zeros((needed_size, self.d_model))
        pe = pe.at[:, 0::2].add(mx.sin(positions * self._div_term))
        pe = pe.at[:, 1::2].add(mx.cos(positions * self._div_term))

        self.pe = pe
        self._pe_cache_len = length

    def __call__(self, x: mx.array) -> Tuple[mx.array, mx.array]:
        """Get positional encodings for sequence.

        Args:
            x: Input tensor (B, T, D)

        Returns:
            Tuple of (scaled_x, pos_emb) where pos_emb has shape (2*T-1, D)
        """
        seq_len = x.shape[1]
        self._extend_pe(seq_len)

        # Scale input if needed
        if self.xscale is not None:
            x = x * self.xscale

        # Return positions for this sequence length
        # For input of length L, we need 2*L-1 positions
        center = self.pe.shape[0] // 2
        start = center - seq_len + 1
        end = center + seq_len
        pos_emb = self.pe[start:end]

        return x, pos_emb


class ConformerFeedForward(nn.Module):
    """Feed-forward module for Conformer."""

    def __init__(
        self,
        d_model: int,
        d_ff: int,
        dropout: float = 0.1,
    ):
        super().__init__()
        self.linear1 = nn.Linear(d_model, d_ff)
        self.linear2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.linear1(x)
        x = nn.silu(x)
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout(x)
        return x


class ConformerConvolution(nn.Module):
    """Convolution module for Conformer."""

    def __init__(
        self,
        d_model: int,
        kernel_size: int = 31,
        norm_type: str = "batch_norm",
        dropout: float = 0.1,
    ):
        super().__init__()
        self.pointwise_conv1 = nn.Linear(d_model, 2 * d_model)

        self.depthwise_conv = nn.Conv1d(
            in_channels=d_model,
            out_channels=d_model,
            kernel_size=kernel_size,
            padding=(kernel_size - 1) // 2,
            groups=d_model,  # Depthwise: each channel processed independently
        )
        if norm_type == "batch_norm":
            self.norm = nn.BatchNorm(d_model)
        else:
            self.norm = nn.LayerNorm(d_model)
        self.pointwise_conv2 = nn.Linear(d_model, d_model)
        self.dropout = nn.Dropout(dropout)

    def __call__(self, x: mx.array) -> mx.array:
        # x: (B, T, D) - already in correct format for MLX
        x = self.pointwise_conv1(x)

        # GLU activation
        x, gate = mx.split(x, 2, axis=-1)
        x = x * mx.sigmoid(gate)

        # Depthwise conv - MLX expects (B, L, C) which is already our format
        x = self.depthwise_conv(x)

        # LayerNorm on features dimension
        x = self.norm(x)

        x = nn.silu(x)
        x = self.pointwise_conv2(x)
        x = self.dropout(x)
        return x


class RelativeMultiHeadAttention(nn.Module):

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dropout: float = 0.1,
        pos_bias: bool = True,
    ):
        super().__init__()
        self.d_model = d_model
        self.num_heads = num_heads
        self.head_dim = d_model // num_heads
        self.scale = 1.0 / math.sqrt(self.head_dim)

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.v_proj = nn.Linear(d_model, d_model)
        self.out_proj = nn.Linear(d_model, d_model)

        self.pos_proj = nn.Linear(d_model, d_model, bias=False)

        if pos_bias:
            self.pos_bias_u = mx.zeros((num_heads, self.head_dim))
            self.pos_bias_v = mx.zeros((num_heads, self.head_dim))
        else:
            self.pos_bias_u = None
            self.pos_bias_v = None

        self.dropout = nn.Dropout(dropout)

    def _rel_shift(self, x: mx.array) -> mx.array:
        """Compute relative positional encoding shift.

        This aligns the position scores so that position i attends to the
        correct relative positions in the positional embedding.

        Args:
            x: Position scores (B, H, T, 2T-1)

        Returns:
            Shifted scores (B, H, T, T) where score[i,j] corresponds to
            relative position i-j.
        """
        B, H, T, pos_len = x.shape
        # Pad on the left: (B, H, T, 2T-1) -> (B, H, T, 2T)
        x = mx.pad(x, [(0, 0), (0, 0), (0, 0), (1, 0)])
        # Reshape: (B, H, T, 2T) -> (B, H, 2T, T)
        x = x.reshape(B, H, pos_len + 1, T)
        # Remove first row: (B, H, 2T, T) -> (B, H, 2T-1, T)
        x = x[:, :, 1:, :]
        # Reshape back: (B, H, 2T-1, T) -> (B, H, T, 2T-1)
        x = x.reshape(B, H, T, pos_len)
        # Take only the first T columns (the valid relative positions)
        return x[:, :, :, :T]

    def __call__(
        self,
        x: mx.array,
        pos_emb: mx.array,
        mask: Optional[mx.array] = None,
    ) -> mx.array:
        """Compute attention with relative positional encoding.

        Args:
            x: Input tensor (B, T, D)
            pos_emb: Positional embeddings (2T-1, D) or (1, 2T-1, D)
            mask: Attention mask

        Returns:
            Output tensor (B, T, D)
        """
        B, T, _ = x.shape

        # Projections: (B, T, D) -> (B, T, H, d_k)
        q = self.q_proj(x).reshape(B, T, self.num_heads, self.head_dim)
        k = self.k_proj(x).reshape(B, T, self.num_heads, self.head_dim)
        v = self.v_proj(x).reshape(B, T, self.num_heads, self.head_dim)

        # Position projection: (2T-1, D) -> (1, 2T-1, H, d_k)
        if pos_emb.ndim == 2:
            pos_emb = pos_emb[None, :, :]
        p = self.pos_proj(pos_emb).reshape(1, -1, self.num_heads, self.head_dim)

        if self.pos_bias_u is not None:
            q_with_bias_u = (q + self.pos_bias_u[None, None, :, :]).transpose(
                0, 2, 1, 3
            )
            q_with_bias_v = (q + self.pos_bias_v[None, None, :, :]).transpose(
                0, 2, 1, 3
            )
        else:
            q_with_bias_u = q.transpose(0, 2, 1, 3)
            q_with_bias_v = q.transpose(0, 2, 1, 3)

        k = k.transpose(0, 2, 1, 3)
        v = v.transpose(0, 2, 1, 3)
        p = p.transpose(0, 2, 1, 3)

        matrix_ac = q_with_bias_u @ k.transpose(0, 1, 3, 2)
        matrix_bd = q_with_bias_v @ p.transpose(0, 1, 3, 2)

        matrix_bd = self._rel_shift(matrix_bd)

        scores = (matrix_ac + matrix_bd) * self.scale

        if mask is not None:
            scores = scores + mask

        attn = mx.softmax(scores, axis=-1)
        attn = self.dropout(attn)

        out = (attn @ v).transpose(0, 2, 1, 3).reshape(B, T, -1)
        return self.out_proj(out)


class ConformerLayer(nn.Module):
    """Single Conformer layer."""

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        ff_expansion_factor: int = 4,
        conv_kernel_size: int = 31,
        conv_norm_type: str = "batch_norm",
        dropout: float = 0.1,
        dropout_att: float = 0.1,
    ):
        super().__init__()
        d_ff = d_model * ff_expansion_factor

        # Pre-norm for each sub-layer
        self.ff1_norm = nn.LayerNorm(d_model)
        self.ff1 = ConformerFeedForward(d_model, d_ff, dropout)

        self.attn_norm = nn.LayerNorm(d_model)
        self.attn = RelativeMultiHeadAttention(d_model, num_heads, dropout_att)

        self.conv_norm = nn.LayerNorm(d_model)
        self.conv = ConformerConvolution(
            d_model, conv_kernel_size, conv_norm_type, dropout
        )

        self.ff2_norm = nn.LayerNorm(d_model)
        self.ff2 = ConformerFeedForward(d_model, d_ff, dropout)

        self.final_norm = nn.LayerNorm(d_model)

    def __call__(
        self,
        x: mx.array,
        pos_emb: mx.array,
        mask: Optional[mx.array] = None,
    ) -> mx.array:
        # First FF (half residual)
        x = x + 0.5 * self.ff1(self.ff1_norm(x))

        # Attention
        x = x + self.attn(self.attn_norm(x), pos_emb, mask)

        # Conv
        x = x + self.conv(self.conv_norm(x))

        # Second FF (half residual)
        x = x + 0.5 * self.ff2(self.ff2_norm(x))

        # Final norm
        x = self.final_norm(x)

        return x


class ConvSubsampling(nn.Module):
    """Convolutional subsampling for audio features using 2D depthwise separable convs."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        subsampling_factor: int = 8,
        conv_channels: int = 256,
        subsampling_type: str = "dw_striding",
    ):
        super().__init__()
        self.subsampling_factor = subsampling_factor
        self.subsampling_type = subsampling_type
        self.conv_channels = conv_channels
        self.in_channels = in_channels

        self.conv = [
            nn.Conv2d(
                1, conv_channels, kernel_size=3, stride=2, padding=1
            ),  # 0: standard conv
            None,  # 1 (ReLU)
            nn.Conv2d(
                conv_channels,
                conv_channels,
                kernel_size=3,
                stride=2,
                padding=1,
                groups=conv_channels,
            ),  # 2
            nn.Conv2d(
                conv_channels, conv_channels, kernel_size=1, stride=1, padding=0
            ),  # 3
            None,  # 4 (ReLU)
            nn.Conv2d(
                conv_channels,
                conv_channels,
                kernel_size=3,
                stride=2,
                padding=1,
                groups=conv_channels,
            ),  # 5
            nn.Conv2d(
                conv_channels, conv_channels, kernel_size=1, stride=1, padding=0
            ),  # 6
        ]

        self.out = nn.Linear(
            conv_channels * (in_channels // subsampling_factor), out_channels
        )

    def __call__(self, x: mx.array) -> mx.array:
        """
        Args:
            x: Input features (B, T, D) where D is mel features

        Returns:
            Subsampled features (B, T', D')
        """
        B, T, D = x.shape

        # Reshape for 2D conv: (B, T, D) -> (B, T, D, 1) - MLX uses NHWC format
        x = x[:, :, :, None]

        x = nn.relu(self.conv[0](x))  # (B, T/2, D/2, 256)

        x = self.conv[2](x)  # (B, T/4, D/4, 256) - depthwise
        x = nn.relu(self.conv[3](x))  # pointwise 1x1 + relu

        x = self.conv[5](x)  # (B, T/8, D/8, 256) - depthwise
        x = nn.relu(self.conv[6](x))  # pointwise 1x1 + relu

        B, T_out, D_out, C = x.shape
        x = x.transpose(0, 1, 3, 2)  # (B, T_out, C, D_out)
        x = x.reshape(B, T_out, -1)  # (B, T_out, C*D_out)
        x = self.out(x)

        return x


class ConformerEncoder(nn.Module):
    """FastConformer encoder for audio processing."""

    def __init__(self, config: ConformerEncoderConfig):
        super().__init__()
        self.config = config

        # Subsampling
        self.pre_encode = ConvSubsampling(
            in_channels=config.feat_in,
            out_channels=config.d_model,
            subsampling_factor=config.subsampling_factor,
            conv_channels=config.subsampling_conv_channels,
            subsampling_type=config.subsampling,
        )

        self.pos_enc = RelativePositionalEncoding(
            config.d_model, config.pos_emb_max_len, xscale=False
        )

        # Pre-encoder dropout
        self.pre_dropout = nn.Dropout(config.dropout_pre_encoder)

        # Conformer layers
        self.layers = [
            ConformerLayer(
                d_model=config.d_model,
                num_heads=config.n_heads,
                ff_expansion_factor=config.ff_expansion_factor,
                conv_kernel_size=config.conv_kernel_size,
                conv_norm_type=config.conv_norm_type,
                dropout=config.dropout,
                dropout_att=config.dropout_att,
            )
            for _ in range(config.n_layers)
        ]

    def __call__(
        self,
        x: mx.array,
        lengths: Optional[mx.array] = None,
    ) -> Tuple[mx.array, mx.array]:
        """
        Args:
            x: Audio features (B, T, D)
            lengths: Original lengths (B,)

        Returns:
            Encoded features (B, T', D) and new lengths
        """
        # Subsampling
        x = self.pre_encode(x)

        # Update lengths
        if lengths is not None:
            lengths = lengths // self.config.subsampling_factor
        else:
            lengths = mx.array([x.shape[1]] * x.shape[0])

        # Get positional embeddings and scale input
        # pos_enc returns (scaled_x, pos_emb) where pos_emb has shape (2T-1, D)
        x, pos_emb = self.pos_enc(x)

        # Pre-encoder dropout
        x = self.pre_dropout(x)

        # Create attention mask if needed
        mask = None
        if lengths is not None:
            max_len = x.shape[1]
            # Create padding mask
            idx = mx.arange(max_len)[None, :]
            mask = idx >= lengths[:, None]
            mask = mx.where(mask[:, None, None, :], float("-inf"), 0.0)

        # Apply conformer layers
        for layer in self.layers:
            x = layer(x, pos_emb, mask)

        return x, lengths


class MLP(nn.Module):
    """MLP adapter for conformer to LFM."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_dims: List[int],
        use_layer_norm: bool = True,
        dropout: float = 0.0,
    ):
        super().__init__()
        channels = [in_channels, *hidden_dims, out_channels]

        layers = []
        if use_layer_norm:
            layers.append(nn.LayerNorm(channels[0]))

        for i in range(len(channels) - 1):
            layers.append(nn.Linear(channels[i], channels[i + 1]))
            if i != len(channels) - 2:
                layers.append(nn.GELU())
                if dropout > 0:
                    layers.append(nn.Dropout(dropout))

        self.layers = layers

    def __call__(self, x: mx.array) -> mx.array:
        for layer in self.layers:
            x = layer(x)
        return x



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import math

import mlx.core as mx
import mlx.nn as nn

from .config import AcousticTokenizerConfig


class ConvRMSNorm(nn.Module):
    """RMSNorm for convolutional features (B, C, T) format."""

    def __init__(self, dim: int, eps: float = 1e-5, elementwise_affine: bool = True):
        super().__init__()
        self.dim = dim
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if elementwise_affine:
            self.weight = mx.ones((dim,))
        else:
            self.weight = None

    def _norm(self, x: mx.array) -> mx.array:
        return x * mx.rsqrt(mx.mean(x**2, axis=-1, keepdims=True) + self.eps)

    def __call__(self, x: mx.array) -> mx.array:
        # x: (B, C, T) -> transpose to (B, T, C) for normalization
        x = mx.transpose(x, (0, 2, 1))
        output = self._norm(x.astype(mx.float32)).astype(x.dtype)
        if self.weight is not None:
            output = output * self.weight
        # Transpose back to (B, C, T)
        return mx.transpose(output, (0, 2, 1))


class CausalConv1d(nn.Module):
    """Causal 1D convolution with padding on the left.

    Input/output format: (B, C, T) - batch, channels, time (PyTorch convention)
    MLX Conv1d expects: (B, T, C) - batch, time, channels
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.groups = groups

        # Calculate padding for causal convolution
        self.padding = (kernel_size - 1) * dilation

        # Use MLX Conv1d with groups parameter
        # For grouped conv, MLX weight shape is (C_out, K, C_in/groups)
        self.conv = nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,  # We handle padding manually
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        # x: (B, C, T) - input in PyTorch format
        # Transpose to MLX format: (B, C, T) -> (B, T, C)
        x = mx.transpose(x, (0, 2, 1))

        # Add causal padding on the time dimension (now axis 1)
        if self.padding > 0:
            x = mx.pad(x, [(0, 0), (self.padding, 0), (0, 0)])

        # Apply conv - MLX expects (B, T, C)
        x = self.conv(x)

        # Transpose back to PyTorch format: (B, T, C) -> (B, C, T)
        x = mx.transpose(x, (0, 2, 1))

        return x


class CausalConvTranspose1d(nn.Module):
    """Causal transposed 1D convolution for upsampling.

    Input/output format: (B, C, T) - batch, channels, time (PyTorch convention)
    MLX ConvTranspose1d expects: (B, T, C) - batch, time, channels
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        bias: bool = True,
        trim_right_ratio: float = 1.0,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.trim_right_ratio = trim_right_ratio

        # Calculate padding
        self.padding_total = kernel_size - stride

        # Use MLX ConvTranspose1d
        self.convtr = nn.ConvTranspose1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        # x: (B, C, T) - input in PyTorch format
        # Transpose to MLX format: (B, C, T) -> (B, T, C)
        x = mx.transpose(x, (0, 2, 1))

        # Apply transposed conv
        x = self.convtr(x)

        # Transpose back to PyTorch format: (B, T, C) -> (B, C, T)
        x = mx.transpose(x, (0, 2, 1))

        # Trim padding for causal (on time dimension, now axis 2)
        padding_right = math.ceil(self.padding_total * self.trim_right_ratio)
        padding_left = self.padding_total - padding_right

        if padding_left > 0:
            x = x[:, :, padding_left:]
        if padding_right > 0:
            x = x[:, :, :-padding_right]

        return x


class DepthwiseConv(nn.Module):
    """Depthwise separable convolution wrapped in a conv module."""

    def __init__(
        self,
        dim: int,
        kernel_size: int = 7,
        causal: bool = True,
        bias: bool = True,
    ):
        super().__init__()
        self.dim = dim
        self.kernel_size = kernel_size
        self.causal = causal

        # Wrapped in another conv module (to match HF structure: mixer.conv.conv.conv)
        self.conv = CausalConv1d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=kernel_size,
            groups=dim,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.conv(x)


class Mixer(nn.Module):
    """Mixer module wrapping depthwise conv."""

    def __init__(
        self, dim: int, kernel_size: int = 7, causal: bool = True, bias: bool = True
    ):
        super().__init__()
        self.conv = DepthwiseConv(dim, kernel_size, causal, bias)

    def __call__(self, x: mx.array) -> mx.array:
        return self.conv(x)


class FeedForward(nn.Module):
    """Feed-forward network with SiLU activation.

    Note: Uses linear1/linear2 naming to match HuggingFace weights.
    """

    def __init__(self, dim: int, mult: float = 4.0, bias: bool = True):
        super().__init__()
        hidden_dim = int(dim * mult)
        self.linear1 = nn.Linear(dim, hidden_dim, bias=bias)
        self.linear2 = nn.Linear(hidden_dim, dim, bias=bias)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.linear1(x)
        x = nn.gelu(x)
        x = self.linear2(x)
        return x


class Block1D(nn.Module):
    """1D convolutional block with depthwise conv and FFN."""

    def __init__(
        self,
        dim: int,
        layernorm: str = "RMSNorm",  # kept for config compatibility
        eps: float = 1e-6,
        causal: bool = True,
        bias: bool = True,
        layer_scale_init_value: float = 1e-6,
    ):
        super().__init__()
        _ = layernorm
        self.dim = dim

        # Normalization
        self.norm = ConvRMSNorm(dim, eps=eps)
        self.ffn_norm = ConvRMSNorm(dim, eps=eps)

        # Mixer (depthwise conv)
        self.mixer = Mixer(dim, kernel_size=7, causal=causal, bias=bias)

        # FFN
        self.ffn = FeedForward(dim, mult=4.0, bias=bias)

        # Layer scale - stored as parameters (gamma, ffn_gamma)
        if layer_scale_init_value > 0:
            self.gamma = mx.ones((dim,)) * layer_scale_init_value
            self.ffn_gamma = mx.ones((dim,)) * layer_scale_init_value
        else:
            self.gamma = None
            self.ffn_gamma = None

    def __call__(self, x: mx.array) -> mx.array:
        # x: (B, C, T)

        # Mixer path
        residual = x
        x = self.norm(x)
        x = self.mixer(x)
        if self.gamma is not None:
            x = x * mx.expand_dims(self.gamma, axis=(0, 2))
        x = residual + x

        # FFN path
        residual = x
        x = self.ffn_norm(x)
        # Transpose for FFN: (B, C, T) -> (B, T, C)
        x = mx.transpose(x, (0, 2, 1))
        x = self.ffn(x)
        # Transpose back: (B, T, C) -> (B, C, T)
        x = mx.transpose(x, (0, 2, 1))
        if self.ffn_gamma is not None:
            x = x * mx.expand_dims(self.ffn_gamma, axis=(0, 2))
        x = residual + x

        return x


class StemConv(nn.Module):
    """Stem convolution layer wrapped in Sequential structure to match HF."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 7,
        bias: bool = True,
    ):
        super().__init__()
        self.conv = CausalConv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.conv(x)


class UpsampleLayer(nn.Module):
    """Upsample layer with transposed convolution."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        bias: bool = True,
    ):
        super().__init__()
        self.convtr = CausalConvTranspose1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.convtr(x)


class HeadConv(nn.Module):
    """Output head convolution."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 7,
        bias: bool = True,
    ):
        super().__init__()
        self.conv = CausalConv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            bias=bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.conv(x)


class TokenizerDecoder(nn.Module):
    """Decoder that converts latent representations back to audio.

    Architecture matches HuggingFace VibeVoice structure:
    - upsample_layers[0] is stem conv
    - upsample_layers[1-6] are transposed convolutions
    - stages[0-6] are transformer blocks
    - head is output convolution
    """

    def __init__(self, config: AcousticTokenizerConfig):
        super().__init__()

        self.dimension = config.vae_dim
        self.channels = config.channels
        self.n_filters = (
            config.decoder_n_filters
            if config.decoder_n_filters
            else config.encoder_n_filters
        )

        # Use decoder ratios or fallback to encoder ratios
        self.ratios = (
            config.decoder_ratios if config.decoder_ratios else config.encoder_ratios
        )

        # Parse depths - should be reversed encoder depths for decoder
        if config.decoder_depths:
            if isinstance(config.decoder_depths, str):
                self.depths = [int(d) for d in config.decoder_depths.split("-")]
            else:
                self.depths = config.decoder_depths
        else:
            if isinstance(config.encoder_depths, str):
                encoder_depths = [int(d) for d in config.encoder_depths.split("-")]
            else:
                encoder_depths = config.encoder_depths
            self.depths = list(reversed(encoder_depths))

        self.causal = config.causal
        self.n_stages = len(self.depths)

        # Upsample layers - wrapped in list structure to match HF naming
        # HF: upsample_layers.X.0.conv or upsample_layers.X.0.convtr
        self.upsample_layers = []

        # First upsample layer is stem conv (upsample_layers.0.0.conv)
        stem_out_ch = self.n_filters * (2 ** (self.n_stages - 1))
        self.upsample_layers.append(
            [
                StemConv(
                    in_channels=self.dimension,
                    out_channels=stem_out_ch,
                    kernel_size=7,
                    bias=config.conv_bias,
                )
            ]
        )

        # Remaining upsample layers are transposed convolutions
        for i in range(len(self.ratios)):
            in_ch = self.n_filters * (2 ** (self.n_stages - 1 - i))
            out_ch = (
                self.n_filters * (2 ** (self.n_stages - 2 - i))
                if i < len(self.ratios) - 1
                else self.n_filters
            )

            self.upsample_layers.append(
                [
                    UpsampleLayer(
                        in_channels=in_ch,
                        out_channels=out_ch,
                        kernel_size=self.ratios[i] * 2,
                        stride=self.ratios[i],
                        bias=config.conv_bias,
                    )
                ]
            )

        # Transformer stages
        self.stages = []
        for i in range(self.n_stages):
            in_ch = self.n_filters * (2 ** (self.n_stages - 1 - i))
            stage_blocks = []
            for _ in range(self.depths[i]):
                stage_blocks.append(
                    Block1D(
                        dim=in_ch,
                        layernorm=config.layernorm,
                        eps=config.layernorm_eps,
                        causal=config.causal,
                        bias=config.conv_bias,
                        layer_scale_init_value=config.layer_scale_init_value,
                    )
                )
            self.stages.append(stage_blocks)

        # Output head
        self.head = HeadConv(
            in_channels=self.n_filters,
            out_channels=config.channels,
            kernel_size=7,
            bias=config.conv_bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        """
        Args:
            x: Latent tensor of shape (B, T, D) or (B, D, T)

        Returns:
            Audio tensor of shape (B, 1, T')
        """
        # Ensure x is in (B, D, T) format
        if x.shape[1] != self.dimension:
            x = mx.transpose(x, (0, 2, 1))

        # Apply stem (first upsample layer)
        x = self.upsample_layers[0][0](x)

        # Process through stages and upsampling
        for i in range(self.n_stages):
            # Apply stage blocks
            for block in self.stages[i]:
                x = block(x)

            # Apply upsampling (skip first upsample which was stem)
            if i + 1 < len(self.upsample_layers):
                x = self.upsample_layers[i + 1][0](x)

        # Output head
        x = self.head(x)

        return x


class AcousticTokenizer(nn.Module):
    """VibeVoice acoustic tokenizer (decoder only for inference)."""

    def __init__(self, config: AcousticTokenizerConfig):
        super().__init__()
        self.config = config
        self.fix_std = config.fix_std
        self.std_dist_type = config.std_dist_type

        self.decoder = TokenizerDecoder(config)

    def decode(self, latents: mx.array) -> mx.array:
        """Convert latent representations to audio.

        Args:
            latents: Latent tensor of shape (B, T, D) where D = vae_dim

        Returns:
            Audio tensor of shape (B, 1, T')
        """
        return self.decoder(latents)

    def __call__(self, latents: mx.array) -> mx.array:
        """Alias for decode."""
        return self.decode(latents)



import math

import mlx.core as mx
import mlx.nn as nn

from .config import FlowLMConfig


def modulate(x: mx.array, shift: mx.array, scale: mx.array) -> mx.array:
    return x * (1 + scale) + shift


def _rms_norm(x: mx.array, alpha: mx.array, eps: float) -> mx.array:
    x_dtype = x.dtype
    x = x.astype(mx.float32)
    var = eps + mx.var(x, axis=-1, keepdims=True, ddof=1)
    y = x * (alpha.astype(var.dtype) * mx.rsqrt(var))
    return y.astype(x_dtype)


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-5):
        super().__init__()
        self.eps = eps
        self.alpha = mx.ones((dim,))

    def __call__(self, x: mx.array) -> mx.array:
        return _rms_norm(x, self.alpha, self.eps)


class LayerNorm(nn.Module):
    """LayerNorm implementation that matches the PyTorch reference."""

    def __init__(
        self, channels: int, eps: float = 1e-6, elementwise_affine: bool = True
    ):
        super().__init__()
        self.eps = eps
        if elementwise_affine:
            self.weight = mx.ones((channels,))
            self.bias = mx.zeros((channels,))
        else:
            self.weight = None
            self.bias = None

    def __call__(self, x: mx.array) -> mx.array:
        return mx.fast.layer_norm(x, self.weight, self.bias, eps=self.eps)


class TimestepEmbedder(nn.Module):
    """Embeds scalar timesteps into vector representations."""

    def __init__(
        self,
        hidden_size: int,
        frequency_embedding_size: int = 256,
        max_period: int = 10000,
    ):
        super().__init__()
        self.frequency_embedding_size = frequency_embedding_size
        half = frequency_embedding_size // 2
        self.freqs = mx.exp(
            -math.log(max_period) * mx.arange(0, half, dtype=mx.float32) / half
        )
        self.mlp = [
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
            RMSNorm(hidden_size),
        ]

    def __call__(self, t: mx.array) -> mx.array:
        if t.ndim == 1:
            t = t[:, None]
        args = t.astype(mx.float32) * self.freqs[None, :]
        embedding = mx.concatenate([mx.cos(args), mx.sin(args)], axis=-1)
        out = embedding
        for layer in self.mlp:
            out = layer(out)
        return out


class ResBlock(nn.Module):
    def __init__(self, channels: int):
        super().__init__()
        self.in_ln = LayerNorm(channels, eps=1e-6)
        self.mlp = [
            nn.Linear(channels, channels, bias=True),
            nn.SiLU(),
            nn.Linear(channels, channels, bias=True),
        ]
        self.adaLN_modulation = [
            nn.SiLU(),
            nn.Linear(channels, 3 * channels, bias=True),
        ]

    def __call__(self, x: mx.array, y: mx.array) -> mx.array:
        modulation = y
        for layer in self.adaLN_modulation:
            modulation = layer(modulation)
        shift_mlp, scale_mlp, gate_mlp = mx.split(modulation, 3, axis=-1)
        h = modulate(self.in_ln(x), shift_mlp, scale_mlp)
        for layer in self.mlp:
            h = layer(h)
        return x + gate_mlp * h


class FinalLayer(nn.Module):
    def __init__(self, model_channels: int, out_channels: int):
        super().__init__()
        self.norm_final = LayerNorm(model_channels, elementwise_affine=False, eps=1e-6)
        self.linear = nn.Linear(model_channels, out_channels, bias=True)
        self.adaLN_modulation = [
            nn.SiLU(),
            nn.Linear(model_channels, 2 * model_channels, bias=True),
        ]

    def __call__(self, x: mx.array, c: mx.array) -> mx.array:
        modulation = c
        for layer in self.adaLN_modulation:
            modulation = layer(modulation)
        shift, scale = mx.split(modulation, 2, axis=-1)
        x = modulate(self.norm_final(x), shift, scale)
        return self.linear(x)


class SimpleMLPAdaLN(nn.Module):
    def __init__(
        self,
        in_channels: int,
        model_channels: int,
        out_channels: int,
        cond_channels: int,
        num_res_blocks: int,
        num_time_conds: int = 1,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.num_time_conds = num_time_conds

        if num_time_conds == 1:
            raise ValueError("num_time_conds must be != 1 for AdaLN conditioning.")

        self.time_embed = [
            TimestepEmbedder(model_channels) for _ in range(num_time_conds)
        ]
        self.cond_embed = nn.Linear(cond_channels, model_channels, bias=True)
        self.input_proj = nn.Linear(in_channels, model_channels, bias=True)
        self.res_blocks = [ResBlock(model_channels) for _ in range(num_res_blocks)]
        self.final_layer = FinalLayer(model_channels, out_channels)

    @classmethod
    def from_pydantic_config(
        cls, cfg: FlowLMConfig, latent_dim: int, cond_dim: int
    ) -> "SimpleMLPAdaLN":
        flow_dim = cfg.flow.dim
        flow_depth = cfg.flow.depth
        num_time_conds = 2
        return cls(
            latent_dim,
            flow_dim,
            latent_dim,
            cond_dim,
            flow_depth,
            num_time_conds=num_time_conds,
        )

    def __call__(self, c: mx.array, s: mx.array, t: mx.array, x: mx.array) -> mx.array:
        ts = [s, t]
        if len(ts) != self.num_time_conds:
            raise ValueError(
                f"Expected {self.num_time_conds} time conditions, got {len(ts)}"
            )
        x = self.input_proj(x)
        t_combined = mx.zeros((x.shape[0], self.model_channels), dtype=x.dtype)
        for idx, time_embed in enumerate(self.time_embed):
            t_combined = t_combined + time_embed(ts[idx])
        t_combined = t_combined / self.num_time_conds
        c = self.cond_embed(c)
        y = t_combined + c
        for block in self.res_blocks:
            x = block(x, y)
        return self.final_layer(x, y)



from typing import List

import mlx.core as mx
import mlx.nn as nn


class OffsetScale(nn.Module):
    """MLX implementation of OffsetScale.

    OffsetScale applies learned offsets and scales to the input tensor for multiple heads.
    It performs element-wise scaling and offset operations and returns a list of tensors,
    one for each head.

    Arguments:
        dim: Dimension of the input features
        heads: Number of heads (default: 1)
    """

    def __init__(self, dim: int, heads: int = 1):
        super().__init__()
        self.dim = dim
        self.heads = heads

        # Initialize scale parameters (gamma) with normal distribution
        self.gamma = mx.random.normal((heads, dim), scale=0.02) + 1.0

        # Initialize offset parameters (beta) with zeros
        self.beta = mx.zeros((heads, dim))

    def __call__(self, x: mx.array) -> List[mx.array]:
        """Forward pass for the OffsetScale layer.

        Args:
            x: Input tensor of shape (..., dim)

        Returns:
            List of tensors with applied offsets and scales for each head.
            Each tensor has shape (..., dim)
        """
        # Apply scaling and offsets using einsum-like operation
        # PyTorch: einsum('... d, h d -> ... h d', x, self.gamma) + self.beta

        # Expand x to include head dimension: (..., dim) -> (..., 1, dim)
        x_expanded = mx.expand_dims(x, axis=-2)

        # Broadcast multiplication with gamma: (..., 1, dim) * (heads, dim) -> (..., heads, dim)
        scaled = x_expanded * self.gamma

        # Add beta offset: (..., heads, dim) + (heads, dim) -> (..., heads, dim)
        out = scaled + self.beta

        # Unbind along the head dimension to create a list of tensors
        # PyTorch: out.unbind(dim=-2)
        head_outputs = []
        for h in range(self.heads):
            head_output = out[..., h, :]  # Extract head h: (..., dim)
            head_outputs.append(head_output)

        return head_outputs



from pathlib import Path
from typing import Optional, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from mlx_audio.utils import base_load_model, get_model_path, load_config

SAMPLE_RATE = 16000

MODEL_REMAPPING = {
    "glm": "glmasr",
    "voxtral": "voxtral",
    "vibevoice": "vibevoice_asr",
    "qwen3_asr": "qwen3_asr",
}


def resample_audio(audio: np.ndarray, orig_sr: int, target_sr: int) -> np.ndarray:
    from scipy import signal

    gcd = np.gcd(orig_sr, target_sr)
    up = target_sr // gcd
    down = orig_sr // gcd
    resampled = signal.resample_poly(audio, up, down, padtype="edge")
    return resampled


def load_audio(
    file: str = Optional[str],
    sr: int = SAMPLE_RATE,
    from_stdin=False,
    dtype: mx.Dtype = mx.float32,
):
    """
    Open an audio file and read as mono waveform, resampling as necessary

    Parameters
    ----------
    file: str
        The audio file to open

    sr: int
        The sample rate to resample the audio if necessary

    Returns
    -------
    A NumPy array containing the audio waveform, in float32 dtype.
    """
    from mlx_audio.audio_io import read as audio_read

    audio, sample_rate = audio_read(file, always_2d=True)
    if sample_rate != sr:
        audio = resample_audio(audio, sample_rate, sr)
    return mx.array(audio, dtype=dtype).mean(axis=1)


def load_model(
    model_path: Union[str, Path], lazy: bool = False, strict: bool = False, **kwargs
) -> nn.Module:
    """
    Load and initialize an STT model from a given path.

    Args:
        model_path: The path or HuggingFace repo to load the model from.
        lazy: If False, evaluate model parameters immediately.
        strict: If True, raise an error if any weights are missing.
        **kwargs: Additional keyword arguments (revision, force_download).

    Returns:
        nn.Module: The loaded and initialized model.
    """
    return base_load_model(
        model_path=model_path,
        category="stt",
        model_remapping=MODEL_REMAPPING,
        lazy=lazy,
        strict=strict,
        **kwargs,
    )


def load(
    model_path: Union[str, Path], lazy: bool = False, strict: bool = False, **kwargs
) -> nn.Module:
    """
    Load a speech-to-text model from a local path or HuggingFace repository.

    This is the main entry point for loading STT models. It automatically
    detects the model type and initializes the appropriate model class.

    Args:
        model_path: The local path or HuggingFace repo ID to load from.
        lazy: If False, evaluate model parameters immediately.
        strict: If True, raise an error if any weights are missing.
        **kwargs: Additional keyword arguments:
            - revision (str): HuggingFace revision/branch to use
            - force_download (bool): Force re-download of model files

    Returns:
        nn.Module: The loaded and initialized model.

    Example:
        >>> from mlx_audio.stt import load
        >>> model = load("mlx-community/whisper-tiny-asr-fp16")
        >>> result = model.generate(audio)
    """
    return load_model(model_path, lazy=lazy, strict=strict, **kwargs)



from .decoder import ConditionalDecoder
from .f0_predictor import ConvRNNF0Predictor
from .flow import CausalMaskedDiffWithXvec
from .flow_matching import CausalConditionalCFM
from .hifigan import HiFTGenerator, Snake
from .mel import mel_spectrogram
from .s3gen import S3Token2Mel, S3Token2Wav
from .xvector import CAMPPlus

__all__ = [
    "HiFTGenerator",
    "Snake",
    "ConvRNNF0Predictor",
    "CAMPPlus",
    "ConditionalDecoder",
    "CausalConditionalCFM",
    "CausalMaskedDiffWithXvec",
    "S3Token2Mel",
    "S3Token2Wav",
    "mel_spectrogram",
]



from dataclasses import dataclass


@dataclass
class VoiceEncConfig:
    """Voice encoder configuration."""

    num_mels: int = 40
    sample_rate: int = 16000
    speaker_embed_size: int = 256
    ve_hidden_size: int = 256
    n_fft: int = 400
    hop_size: int = 160
    win_size: int = 400
    fmax: int = 8000
    fmin: int = 0
    preemphasis: float = 0.0
    mel_power: float = 2.0
    mel_type: str = "amp"
    normalized_mels: bool = False
    ve_partial_frames: int = 160
    ve_final_relu: bool = True
    stft_magnitude_min: float = 1e-4



# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

from dataclasses import dataclass
from typing import Optional

# GPT2 Medium configuration for Turbo model
GPT2_MEDIUM_CONFIG = {
    "activation_function": "gelu_new",
    "n_ctx": 8196,
    "n_embd": 1024,
    "hidden_size": 1024,
    "n_head": 16,
    "n_layer": 24,
    "n_positions": 8196,
    "vocab_size": 50276,
    "layer_norm_epsilon": 1e-05,
    "attn_pdrop": 0.1,
    "embd_pdrop": 0.1,
    "resid_pdrop": 0.1,
}


@dataclass
class T3Config:
    """Configuration for T3 TTS model."""

    # Text tokens
    start_text_token: int = 255
    stop_text_token: int = 0
    text_tokens_dict_size: int = 50276
    max_text_tokens: int = 2048

    # Speech tokens
    start_speech_token: int = 6561
    stop_speech_token: int = 6562
    speech_tokens_dict_size: int = 6563
    max_speech_tokens: int = 4096

    # Model architecture
    llama_config_name: str = "GPT2_medium"
    input_pos_emb: Optional[str] = None
    speech_cond_prompt_len: int = 375

    # Conditioning
    encoder_type: str = "voice_encoder"
    speaker_embed_size: int = 256
    use_perceiver_resampler: bool = False
    emotion_adv: bool = False

    @property
    def n_channels(self) -> int:
        """Get hidden size from config."""
        return GPT2_MEDIUM_CONFIG["hidden_size"]

    @property
    def is_multilingual(self) -> bool:
        """Check if this is a multilingual model."""
        return self.text_tokens_dict_size == 2454

    @classmethod
    def turbo(cls) -> "T3Config":
        """Create configuration for Turbo TTS model."""
        return cls(
            text_tokens_dict_size=50276,
            speech_tokens_dict_size=6563,
            llama_config_name="GPT2_medium",
            input_pos_emb=None,
            speech_cond_prompt_len=375,
            use_perceiver_resampler=False,
            emotion_adv=False,
        )
